"""Implementation of add unit subcommand"""

from twisted.internet.defer import inlineCallbacks

from ensemble.state.errors import NoUnusedMachines
from ensemble.state.service import ServiceStateManager
from ensemble.state.machine import MachineStateManager
from ensemble.control.utils import get_environment


def configure_subparser(subparsers):
    """Configure add-unit subcommand"""
    sub_parser = subparsers.add_parser("add-unit", help=command.__doc__)
    sub_parser.add_argument(
        "--environment", "-e",
        help="Ensemble environment to operate in.")
    sub_parser.add_argument(
        "service_name",
        help="Name of the service a unit should be added for")
    return sub_parser


def command(options):
    """Add a new service unit."""
    environment = get_environment(options)
    return add_unit(
        options.environments,
        environment,
        options.verbose,
        options.log,
        options.service_name)


@inlineCallbacks
def add_unit(config, environment, verbose, log, service_name):
    provider = environment.get_machine_provider()
    client = yield provider.connect()
    try:
        service_manager = ServiceStateManager(client)
        machine_manager = MachineStateManager(client)

        service_state = yield service_manager.get_service_state(service_name)
        unit_state = yield service_state.add_unit_state()
        try:
            yield unit_state.assign_to_unused_machine()
        except NoUnusedMachines:
            machine_state = yield machine_manager.add_machine_state()
            yield unit_state.assign_to_machine(machine_state)
        log.info("Unit %r added to service %r",
            unit_state.unit_name, service_state.service_name)
    finally:
        yield client.close()
