"""
Command for debugging hooks on a service unit.
"""
import base64
import os

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.control.ssh import get_ip_address_for_unit
from ensemble.control.utils import get_environment

from ensemble.formula.errors import InvalidFormulaHook
from ensemble.state.formula import FormulaStateManager
from ensemble.state.service import ServiceStateManager


def configure_subparser(subparsers):
    sub_parser = subparsers.add_parser("debug-hooks", help=command.__doc__)
    sub_parser.add_argument(
        "-e", "--environment",
        help="Ensemble environment to operate in.")

    sub_parser.add_argument(
        "unit_name",
        help="Name of unit")

    sub_parser.add_argument(
        "hook_names", default=["*"], nargs="*",
        help="Name of hook, defaults to all")

    return sub_parser


@inlineCallbacks
def validate_hooks(client, unit_state, hook_names):

    # Assemble a list of valid hooks for the formula.
    valid_hooks = ["start", "stop", "install", "config-changed"]
    service_manager = ServiceStateManager(client)
    endpoints = yield service_manager.get_relation_endpoints(
        unit_state.service_name)
    endpoint_names = [ep.relation_name for ep in endpoints]
    for endpoint_name in endpoint_names:
        valid_hooks.extend([
            endpoint_name + "-relation-joined",
            endpoint_name + "-relation-changed",
            endpoint_name + "-relation-departed",
            endpoint_name + "-relation-broken",
        ])

    # Verify the debug names.
    for hook_name in hook_names:
        if hook_name in valid_hooks:
            continue
        break
    else:
        returnValue(True)

    # We dereference to the formula to give a fully qualified error
    # message.  I wish this was a little easier to dereference, the
    # service_manager.get_relation_endpoints effectively does this
    # already.
    service_manager = ServiceStateManager(client)
    service_state = yield service_manager.get_service_state(
        unit_state.service_name)
    formula_id = yield service_state.get_formula_id()
    formula_manager = FormulaStateManager(client)
    formula = yield formula_manager.get_formula_state(formula_id)
    raise InvalidFormulaHook(formula.id, hook_name)


@inlineCallbacks
def command(options):
    """Interactively debug a hook remotely on a service unit.
    """
    environment = get_environment(options)
    provider = environment.get_machine_provider()
    client = yield provider.connect()

    options.log.debug("Retrieving unit and machine information.")
    ip_address, unit, machine = yield get_ip_address_for_unit(
        client, provider, options.unit_name)

    # Verify hook name
    if options.hook_names != ["*"]:
        options.log.debug("Verifying hook names...")
        yield validate_hooks(client, unit, options.hook_names)

    if not ip_address:
        options.log.info("Remote machine not available")
        yield client.close()
        returnValue(True)

    options.log.debug(
        "Enabling hook debug on unit (%r)..." % options.unit_name)
    yield unit.enable_hook_debug(options.hook_names)

    # Connect via ssh and start tmux.
    options.log.info("Connecting to remote machine...")

    # Encode the script as base64 so that we can deliver it with a single
    # ssh command while still retaining standard input on the terminal fd.
    script = SCRIPT.replace("{unit_name}", options.unit_name)
    script_b64 = base64.encodestring(script).replace("\n", "").strip()
    cmd = '"F=`mktemp`; echo %s | base64 -d > \$F; . \$F"' % script_b64

    # Yield to facilitate testing.
    yield os.system("ssh -t ubuntu@%s 'sudo /bin/bash -c %s'" % (ip_address, cmd))

    options.log.info("Debug session ended.")
    # Ends hook debugging.
    yield client.close()


SCRIPT = r"""
# Wait for tmux to be installed.
while [ ! -f /usr/bin/tmux ]; do
    sleep 1
done

# Create a sane configuration (no green bars, please!)
[ -f ~/.tmux.conf ] || cat > ~/.tmux.conf <<END

# Status bar
set-option -g status-bg black
set-option -g status-fg white

set-window-option -g window-status-current-bg red
set-window-option -g window-status-current-attr bright

set-option -g status-right ''

# Panes
set-option -g pane-border-fg white
set-option -g pane-active-border-fg white

# Monitor activity on windows
set-window-option -g monitor-activity on

# Screen bindings, since people are more familiar with that.
set-option -g prefix C-a
bind C-a last-window
bind a send-key C-a

bind | split-window -h
bind - split-window -v

# Fix CTRL-PGUP/PGDOWN for vim
set-window-option -g xterm-keys on

# Prevent ESC key from adding delay and breaking Vim's ESC > arrow key
set-option -s escape-time 0

END

tmux new-session -d -s {unit_name} 2> /dev/null || true
tmux attach -t {unit_name}
"""
