import argparse
import os

import yaml

from twisted.internet.defer import inlineCallbacks

from ensemble.control.utils import get_environment

from ensemble.formula.errors import ServiceConfigError
from ensemble.formula.publisher import FormulaPublisher
from ensemble.formula.repository import LocalFormulaRepository

from ensemble.state.endpoint import RelationEndpoint
from ensemble.state.environment import EnvironmentStateManager
from ensemble.state.errors import NoUnusedMachines
from ensemble.state.machine import MachineStateManager
from ensemble.state.relation import RelationStateManager
from ensemble.state.service import ServiceStateManager


def configure_subparser(subparsers):
    sub_parser = subparsers.add_parser("deploy", help=command.__doc__)

    sub_parser.add_argument(
        "--environment", "-e",
        help="Environment to deploy the formula in.")

    sub_parser.add_argument(
        "--repository", required=True,
        help="Directory for formula lookup and retrieval",
        type=os.path.abspath)

    sub_parser.add_argument(
        "formula", nargs=None,
        help="Formula name")

    sub_parser.add_argument(
        "service_name", nargs="?",
        help="Service name of deployed formula")

    sub_parser.add_argument(
        "--config",
        help="YAML file containing service options")



    return sub_parser


def command(options):
    """
    Deploy a formula to Ensemble!
    """
    environment = get_environment(options)
    return deploy(
        options.environments,
        environment,
        options.repository,
        options.formula,
        options.service_name,
        options.log,
        options.config)

def parse_config_options(config_file, service_name):
    if not os.path.exists(config_file) or \
            not os.access(config_file, os.R_OK):
        raise ServiceConfigError("Config file '%s' not accessible." % config_file)

    options = yaml.load(open(config_file, "r").read())
    if not options or not isinstance(options, dict) or \
            service_name not in options:
        raise ServiceConfigError(
            "Invalid options file passed to --config.\n"
            "Expected a YAML dict with service name (%r)." % service_name)
    # remove the service name prefix for application to state
    return options[service_name]



@inlineCallbacks
def deploy(env_config, environment, repository_path, formula_name,
           service_name, log, config_file=None):
    """Deploy a formula within an environment.

    This will publish the formula to the environment, creating
    a service from the formula, and get it set to be launched
    on a new machine.
    """
    service_options = None
    service_name = service_name or formula_name

    # Validate config options prior to deployment attempt
    if config_file:
        service_options = parse_config_options(config_file, service_name)

    repository = LocalFormulaRepository(repository_path)
    formula = repository.find(formula_name)

    provider = environment.get_machine_provider()
    client = yield provider.connect()
    try:
        storage = yield provider.get_file_storage()
        service_manager = ServiceStateManager(client)
        machine_manager = MachineStateManager(client)
        environment_state_manager = EnvironmentStateManager(client)
        yield environment_state_manager.set_config_state(
            env_config, environment.name)

        # Publish the formula to ensemble
        publisher = FormulaPublisher(client, storage)
        yield publisher.add_formula("local", formula)
        result = yield publisher.publish()

        # In future we might have multiple formulas be published at
        # the same time.  For now, extract the formula_state from the
        # list.
        formula_state = result[0]

        # Create the service state
        service_state = yield service_manager.add_service_state(
            service_name, formula_state)

        # now if config options were passed make those available
        if service_options:
            state = yield service_state.get_config()
            state.update(service_options)
            yield state.write()

        # Create a service unit
        unit_state = yield service_state.add_unit_state()

        try:
            # Attempt to assign to an unused machine
            yield unit_state.assign_to_unused_machine()
        except NoUnusedMachines:
            # Otherwise create the machine state
            machine_state = yield machine_manager.add_machine_state()
            # Associate the unit to the machine
            yield unit_state.assign_to_machine(machine_state)

        # Check if we have any peer relations to establish
        if formula.metadata.peers:
            relation_manager = RelationStateManager(client)
            for peer_name, peer_info in formula.metadata.peers.items():
                yield relation_manager.add_relation_state(
                    RelationEndpoint(service_name,
                                     peer_info["interface"],
                                     peer_name,
                                     "peer"))

        log.info("Formula deployed as service: %r", service_name)
    finally:
        yield client.close()
