import os

import yaml

from ensemble.errors import (
    EnsembleError, InvalidEnsembleHeaderValue, FileNotFound)
from ensemble.formula.errors import FormulaIdError
from ensemble.lib.schema import (
    SchemaError, Bool, Constant, Dict, Int,
    KeyDict, OneOf, UnicodeOrString)


class MetaDataError(EnsembleError):
    """Raised when an error in the info file of a formula is found."""


def parse_formula_id(formula_id):
    """Parse a formula id into its components.

    A formula id is a qualified reference to a specific formula
    from a specific origin. It is used in several places within
    ensemble to refer unambigiously to a formula.

    Formula ids take the form of a string with the following syntax::

       namespace::formula_name-revision

    The returned tuple contains values for namespace, name, and
    revision in that order ie::

       ("local", "mysql", 21)

    """

    if not isinstance(formula_id, basestring):
        raise FormulaIdError(formula_id, "wrong type")

    if not formula_id.count(":") >= 1:
        raise FormulaIdError(formula_id, "invalid namespace")

    namespace, formula_qualifier = formula_id.split(":", 1)

    if not formula_qualifier.count("-") >= 1:
        raise FormulaIdError(formula_id, "invalid revision")

    name, revision = formula_qualifier.rsplit("-", 1)
    if not revision.isdigit():
        raise FormulaIdError(formula_id, "invalid revision")

    return (namespace, name, int(revision))


INTERFACE_SCHEMA = KeyDict({
        "interface": UnicodeOrString("utf-8"),
        "limit": OneOf(Constant(None), Int()),
        "optional": Bool()})


class ReshapeInterfaceSpec(object):
    """Support reshaping of a YAML interface spec to list of interfaces.

    We need this class because our formula shorthand is difficult to
    work with (unfortunately). So we coerce shorthand and then store
    the desired format in ZK.

    Supports the following variants::

      provides:
        server: riak
        admin: http
        foobar:
          interface: blah

      provides:
        server:
          interface: mysql
          limit:
          optional: false

    In all cases, this is converted to a fully specified
    representation, as seen in the mysql interface description above
    (in YAML format)

    KT - Its really not obvious from the doc strings, variable names,
    or implementation what this class is doing.
    """

    value_converter = UnicodeOrString("utf-8")

    def __init__(self, limit):
        """Create relation interface reshaper.

        @limit: the limit for this relation. Used to provide defaults
            for a given kind of relation role (peer, provider, consumer)
        """
        self.limit = limit

    def coerce_interface(self, interface_spec, path):
        """Coerce `interface_spec` from shorthand form.

        Helper method to support each of the variants, either the
        formula does not specify limit and optional, such as foobar in
        the above example; or the interface spec is just a string,
        such as the ``server: riak`` example.
        """
        if not isinstance(interface_spec, dict):
            return {
                "interface":
                    self.value_converter.coerce(interface_spec, path),
                "limit": self.limit,
                "optional": False}
        else:
            # Optional values are context-sensitive and/or have
            # defaults, which is different than what KeyDict can
            # readily support. So just do it here first, then
            # coerce.
            if "limit" not in interface_spec:
                interface_spec["limit"] = self.limit
            if "optional" not in interface_spec:
                interface_spec["optional"] = False
            return INTERFACE_SCHEMA.coerce(interface_spec, path)

    def coerce(self, value, path):
        """Converts an interface specification in shorthand form.

        @value: the value to be coerced
        @path: used in exception reporting"""
        if not isinstance(value, dict):
            raise SchemaError(path, "Invalid relation specification")

        return dict(
            (self.value_converter.coerce(K, path),
             self.coerce_interface(V, path))
            for K, V in value.iteritems())


RELATION_SCHEMA = Dict(
    UnicodeOrString("utf-8"), INTERFACE_SCHEMA)

SCHEMA = KeyDict({
    "ensemble": Constant("formula"),
    "name": UnicodeOrString("utf-8"),
    "revision": Int(),
    "summary": UnicodeOrString("utf-8"),
    "description": UnicodeOrString("utf-8"),
    "peers": OneOf(RELATION_SCHEMA,
                   ReshapeInterfaceSpec(limit=1)),
    "provides": OneOf(RELATION_SCHEMA,
                      ReshapeInterfaceSpec(limit=None)),
    "requires": OneOf(RELATION_SCHEMA,
                      ReshapeInterfaceSpec(limit=1)),
    }, optional=set(["provides", "requires", "peers"]))


class MetaData(object):
    """Represents the formula info file.

    The main metadata for a formula (name, revision, etc) is maintained
    in the formula's info file.  This class is able to parse,
    validate, and provide access to data in the info file.
    """

    def __init__(self, path=None):
        self._data = {}
        if path is not None:
            self.load(path)

    @property
    def name(self):
        """The formula name."""
        return self._data.get("name")

    @property
    def revision(self):
        """The formula revision.

        The formula revision acts as a version, but unlike e.g. package
        versions, the formula revision is a monotonically increasing
        integer.
        """
        return self._data.get("revision")

    @property
    def summary(self):
        """The formula summary."""
        return self._data.get("summary")

    @property
    def description(self):
        """The formula description."""
        return self._data.get("description")

    @property
    def provides(self):
        """The formula provides relations."""
        return self._data.get("provides")

    @property
    def requires(self):
        """The formula requires relations."""
        return self._data.get("requires")

    @property
    def peers(self):
        """The formula peers relations."""
        return self._data.get("peers")

    def get_serialization_data(self):
        """Get internal dictionary representing the state of this instance.

        This is useful to embed this information inside other storage-related
        dictionaries.
        """
        return dict(self._data)

    def get_formula_id(self, namespace):
        """
        Construct a formula id given a namespace.
        @param namespace: A string representing the namespace the formula
                          origin.
        """
        if not isinstance(namespace, basestring) or ":" in namespace:
            raise AssertionError("Invalid namespace string %r" % namespace)
        return "%s:%s-%s" % (namespace, self.name, self.revision)

    def load(self, path):
        """Load and parse the info file.

        @param path: Path of the file to load.

        Internally, this function will pass the content of the file to
        the C{parse()} method.
        """
        if not os.path.isfile(path):
            raise FileNotFound(path)
        with open(path) as f:
            self.parse(f.read(), path)

    def parse(self, content, path=None):
        """Parse the info file described by the given content.

        @param content: Content of the info file to parse.
        @param path: Optional path of the loaded file.  Used when raising
            errors.

        @raise MetaDataError: When errors are found in the info data.
        """
        return self.parse_serialization_data(yaml.load(content), path)

    def parse_serialization_data(self, serialization_data, path=None):
        """Parse the unprocessed serialization data and load in this instance.

        @param serialization_data: Unprocessed data matching the
            metadata schema.
        @param path: Optional path of the loaded file.  Used when
            raising errors.

        @raise MetaDataError: When errors are found in the info data.
        """
        expected_value = "formula"
        found_value = serialization_data.get("ensemble")
        if found_value != expected_value:
            raise InvalidEnsembleHeaderValue(path, expected_value, found_value)
        try:
            self._data = SCHEMA.coerce(serialization_data, [])
        except SchemaError, error:
            if path:
                path_info = " %s:" % path
            else:
                path_info = ""
            raise MetaDataError("Bad data in formula info:%s %s" %
                                (path_info, error))
