# -*- encoding: utf-8 -*-

import os
import yaml
import inspect

from ensemble.lib.testing import TestCase
from ensemble.formula.metadata import (
    MetaData, MetaDataError, ReshapeInterfaceSpec, SchemaError,
    parse_formula_id)
from ensemble.errors import InvalidEnsembleHeaderValue, FileNotFound
from ensemble.formula.errors import FormulaIdError
from ensemble.formula import tests

test_repository_path = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)),
    "repository")
sample_path = os.path.join(test_repository_path, "dummy", "metadata.yaml")
sample_configuration = open(sample_path).read()


class MetaDataTest(TestCase):

    def setUp(self):
        self.metadata = MetaData()
        self.sample = sample_configuration

    def change_sample(self):
        """Return a context manager for hacking the sample data.

        This should be used follows:

            with self.change_sample() as data:
                data["some-key"] = "some-data"

        The changed sample file content will be available in self.sample
        once the context is done executing.
        """

        class HackManager(object):

            def __enter__(mgr):
                mgr.data = yaml.load(self.sample)
                return mgr.data

            def __exit__(mgr, exc_type, exc_val, exc_tb):
                self.sample = yaml.dump(mgr.data)
                return False
        return HackManager()

    def test_path_argument_loads_formula_info(self):
        info = MetaData(sample_path)
        self.assertEquals(info.name, "dummy")

    def test_parse_formula_id(self):
        """Parse a formula id reference into its component parts.

        Also includes validation of the formula id value.
        """
        self.assertRaises(FormulaIdError, parse_formula_id, None)
        self.assertRaises(FormulaIdError, parse_formula_id, 21)
        self.assertRaises(FormulaIdError, parse_formula_id, "abc")
        self.assertRaises(FormulaIdError, parse_formula_id, "local:x-abc")

        self.assertEqual(
            parse_formula_id("local:mysql-21"),
            ("local", "mysql", 21))

        self.assertEqual(
            parse_formula_id("local:mysql-cluster-21"),
            ("local", "mysql-cluster", 21))

    def test_check_basic_info_before_loading(self):
        """
        Attributes should be set to None before anything is loaded.
        """
        self.assertEquals(self.metadata.name, None)
        self.assertEquals(self.metadata.revision, None)
        self.assertEquals(self.metadata.summary, None)
        self.assertEquals(self.metadata.description, None)

    def test_parse_and_check_basic_info(self):
        """
        Parsing the content file should work. :-)  Basic information will
        be available as attributes of the info file.
        """
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.name, "dummy")
        self.assertEquals(self.metadata.revision, 1)
        self.assertEquals(self.metadata.summary, u"That's a dummy formula.")
        self.assertEquals(self.metadata.description,
                          u"This is a longer description which\n"
                          u"potentially contains multiple lines.\n")

    def test_load_calls_parse_calls_parse_serialzation_data(self):
        """
        We'll break the rules a little bit here and test the implementation
        itself just so that we don't have to test *everything* twice. If
        load() calls parse() which calls parse_serialzation_data(), then
        whatever happens with parse_serialization_data(), happens with the
        others.
        """
        serialization_data = {"Hi": "there!"}
        yaml_data = yaml.dump(serialization_data)
        path = self.makeFile(yaml_data)
        mock = self.mocker.patch(self.metadata)
        mock.parse(yaml_data, path)
        self.mocker.passthrough()
        mock.parse_serialization_data(serialization_data, path)
        self.mocker.replay()

        self.metadata.load(path)
        # Do your magic Mocker!

    def test_needs_ensemble_header(self):
        """
        The loaded content should have a "ensemble: formula" header.
        """
        with self.change_sample() as data:
            data["ensemble"] = "foo"
        error = self.assertRaises(InvalidEnsembleHeaderValue,
                                  self.metadata.parse, self.sample)
        self.assertEquals(error.expected, "formula")
        self.assertEquals(error.found, "foo")
        self.assertEquals(error.path, None)

    def test_needs_ensemble_header_with_load(self):
        """
        Check that the exception raised by load() on header errors will
        include the correct path in the exception, so that the user can
        tell where the issue is.
        """
        with self.change_sample() as data:
            data["ensemble"] = "foo"
        filename = self.makeFile(self.sample)
        error = self.assertRaises(InvalidEnsembleHeaderValue,
                                  self.metadata.load, filename)
        self.assertEquals(error.expected, "formula")
        self.assertEquals(error.found, "foo")
        # That's why we need this test:
        self.assertEquals(error.path, filename)

    def test_schema_error_includes_path_with_load(self):
        """
        When using load(), the exception message should mention the
        path name which was attempted.
        """
        with self.change_sample() as data:
            data["revision"] = "1"
        filename = self.makeFile(self.sample)
        error = self.assertRaises(MetaDataError,
                                  self.metadata.load, filename)
        self.assertEquals(str(error),
                          "Bad data in formula info: %s: revision: "
                          "expected int, got '1'" % filename)

    def test_load_missing_file(self):
        """
        When using load(), the exception message should mention the
        path name which was attempted.
        """
        filename = self.makeFile()
        error = self.assertRaises(FileNotFound,
                                  self.metadata.load, filename)
        self.assertEquals(error.path, filename)

    def test_revision_is_int(self):
        """
        Revision numbers are *integers*. Yep, no 1.2.3 versioning.
        """
        with self.change_sample() as data:
            data["revision"] = "1"
        error = self.assertRaises(MetaDataError,
                                  self.metadata.parse, self.sample)
        self.assertEquals(str(error),
                          "Bad data in formula info: revision: "
                          "expected int, got '1'")

    def test_name_summary_and_description_are_utf8(self):
        """
        Textual fields are decoded to unicode by the schema using UTF-8.
        """
        value = u"áéíóú"
        str_value = value.encode("utf-8")
        with self.change_sample() as data:
            data["name"] = str_value
            data["summary"] = str_value
            data["description"] = str_value
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.name, value)
        self.assertEquals(self.metadata.summary, value)
        self.assertEquals(self.metadata.description, value)

    def test_get_serialized_data(self):
        """
        The get_serialization_data() function should return an object which
        may be passed to parse_serialization_data() to restore the state of
        the instance.
        """
        self.metadata.parse(self.sample)
        serialization_data = self.metadata.get_serialization_data()
        self.assertEquals(serialization_data["name"], "dummy")

    def test_get_formula_id(self):
        """Metadata should be able to construct a formula id given a namespace.
        """
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.get_formula_id("local"),
                          "local:%s-%s" % (
                              self.metadata.name, self.metadata.revision))

    def test_get_formula_id_invalid_namespace(self):
        """An invalid namespace is detected and raises an error
        """
        self.metadata.parse(self.sample)
        self.assertRaises(AssertionError, self.metadata.get_formula_id, None)
        self.assertRaises(AssertionError, self.metadata.get_formula_id, 1)
        self.assertRaises(AssertionError, self.metadata.get_formula_id, "z:x")


class ParseTest(TestCase):
    """Test the parsing of some well-known sample files"""

    def get_metadata(self, formula_name):
        """Get the associated metadata for a given formula, eg ``wordpress``"""
        metadata = MetaData(os.path.join(
            test_repository_path, formula_name, "metadata.yaml"))
        self.assertEqual(metadata.name, formula_name)
        return metadata

    def test_mysql_sample(self):
        """Test parse of a relation written in shorthand format.

        Such relations are defined as follows::
           provides:
             server: mysql
        """
        metadata = self.get_metadata("mysql")
        self.assertEqual(metadata.peers, None)
        self.assertEqual(
            metadata.provides["server"],
            {"interface": "mysql", "limit": None, "optional": False})
        self.assertEqual(metadata.requires, None)

    def test_riak_sample(self):
        """Test multiple interfaces defined in long form, with defaults."""
        metadata = self.get_metadata("riak")
        self.assertEqual(
            metadata.peers["ring"],
            {"interface": "riak", "limit": 1, "optional": False})
        self.assertEqual(
            metadata.provides["endpoint"],
            {"interface": "http", "limit": None, "optional": False})
        self.assertEqual(
            metadata.provides["admin"],
            {"interface": "http", "limit": None, "optional": False})
        self.assertEqual(metadata.requires, None)

    def test_wordpress_sample(self):
        """Test multiple interfaces defined in long form, without defaults."""
        metadata = self.get_metadata("wordpress")
        self.assertEqual(metadata.peers, None)
        self.assertEqual(
            metadata.provides["url"],
            {"interface": "http", "limit": None, "optional": False})
        self.assertEqual(
            metadata.requires["db"],
            {"interface": "mysql", "limit": 1, "optional": False})
        self.assertEqual(
            metadata.requires["cache"],
            {"interface": "varnish", "limit": 2, "optional": True})

    def test_reshape_interface_spec(self):
        """Test rewriting of a given interface specification into long form.

        ReshapeInterfaceSpec uses `coerce` to do one of two things:

          - Rewrite shorthand to the long form used for actual storage
          - Fills in defaults, including a configurable `limit`

        This test ensures test coverage on each of these branches, along
        with ensuring the conversion object properly raises SchemaError
        exceptions on invalid data.
        """
        provides_reshaper = ReshapeInterfaceSpec(limit=None)

        # shorthand is properly rewritten
        self.assertEqual(
            provides_reshaper.coerce({"admin": "http"}, ["provides"]),
            {"admin": {"interface": "http", "limit": None, "optional": False}})

        # defaults are properly applied
        self.assertEqual(
            provides_reshaper.coerce(
                {"admin": {"interface": "http"}}, ["provides"]),
            {"admin": {"interface": "http", "limit": None, "optional": False}})
        self.assertEqual(
            provides_reshaper.coerce(
                {"foobar": {"interface": "http", "limit": 2}}, ["provides"]),
            {"foobar": {"interface": "http", "limit": 2, "optional": False}})
        self.assertEqual(
            provides_reshaper.coerce(
                {"foobar": {"interface": "http", "optional": True}},
                ["provides"]),
            {"foobar": {"interface": "http", "limit": None, "optional": True}})

        # invalid data raises SchemaError
        self.assertRaises(
            SchemaError,
            provides_reshaper.coerce, {"admin": 42}, ["provides"])
        self.assertRaises(
            SchemaError,
            provides_reshaper.coerce,
            {"admin": {"interface": "http", "optional": None}},
            ["provides"])
        self.assertRaises(
            SchemaError,
            provides_reshaper.coerce,
            {"admin": {"interface": "http", "limit": "none, really"}},
            ["provides"])

        # can change `limit` default
        consumer_reshaper = ReshapeInterfaceSpec(limit=1)
        self.assertEqual(
            consumer_reshaper.coerce({"admin": "http"}, ["consumes"]),
            {"admin": {"interface": "http", "limit": 1, "optional": False}})
