import os
import sys

from twisted.internet import protocol, defer
from twisted.python.failure import Failure

from ensemble import errors


class HookProtocol(protocol.ProcessProtocol):
    """
    Protocol used to communicate between the Unit Agent and a
    hook. This serves as a message router from POSIX process status
    and streams and the invoking process.

    It passes process exit status back to the controlling Invoker via
    a deferred.
    """
    def __init__(self, hook_name, context, log=None):
        self._context = context
        self._hook_name = hook_name
        self._log = log
        self.deferred = defer.Deferred()

    def outReceived(self, data):
        self._log.info(data)

    def errReceived(self, data):
        self._log.error(data)

    def processEnded(self, reason):
        exit_code = reason.value.exitCode
        if exit_code in (0, None):
            self.deferred.callback(exit_code)
        else:
            error = errors.FormulaInvocationError(self._hook_name, exit_code)
            self.deferred.errback(error)


class Invoker(object):
    """Responsible for the execution and management of hook
    invocation. 'How' hooks are invoked, not when or why. This manages
    the socket connection from the Unit Agent, connecting standard
    descriptors to logging and handing back exit codes via a Deferred.

    arguments:

    `context` -- an ensemble.state.hook.HookContext

    `change` -- an ensemble.state.hook.RelationChange

    `client_id` -- a string uniquely identifying a client session

    `socket_path` -- the path to the UNIX Domain socket used by
                     clients to communicate with the Unit Agent

    `logger` -- instance of a logging.Logger object used to capture
                hook output.
    """
    def __init__(self, context, change, client_id, socket_path,
                 unit_path, logger):
        self.environment = {}
        self._context = context
        self._change = change
        self._client_id = client_id
        self._socket_path = socket_path
        self._unit_path = unit_path
        self._log = logger
        # The twisted.internet.process.Process instance.
        self._process = None
        # The hook executable path
        self._process_executable = None

    @property
    def unit_path(self):
        return self._unit_path

    def get_environment(self):
        """
        Returns the environment used to run the hook as a dict.
        Defaults are provided based on information passed to __init__.
        By setting keys inside Invoker.environment you can override
        the defaults or provide additional variables.
        """
        base = dict(ENSEMBLE_AGENT_SOCKET=self._socket_path,
                    ENSEMBLE_CLIENT_ID=self._client_id,
                    FORMULA_DIR=os.path.join(self._unit_path, "formula"),
                    ENSEMBLE_UNIT_NAME=os.environ["ENSEMBLE_UNIT_NAME"],
                    PATH=os.environ["PATH"],
                    PYTHON=os.environ.get("PYTHON", sys.executable),
                    PYTHONPATH=":".join(sys.path))
        base.update(self.environment)

        return self.get_environment_from_change(base, self._change)

    def get_environment_from_change(self, env, change):
        """Supplement the default environment with dict with variables
        originating from the `change` argument to __init__.
        """
        return env

    def get_context(self):
        """Returns the hook context for the invocation."""
        return self._context

    def validate_hook(self, hook_filename):
        """ Verify that the hook_filename exists and is executable. """
        if not os.path.exists(hook_filename):
            raise errors.FileNotFound(hook_filename)

        if not os.access(hook_filename, os.X_OK):
            raise errors.FormulaError(hook_filename,
                                      "hook is not executable")

    def send_signal(self, signal_id):
        """Send a signal of the given signal_id.

        :param signal_id: limited value interpretation, numeric
        signals ids are used as given, some values for symbolic
        string interpretation are available see
        ``twisted.internet.process._BaseProcess.signalProcess`` for
        additional details.

        Raises a ValueError if the process doesn't exist or
        ProcessExitedAlreaday if the process has already ended.
        """

        if not self._process:
            raise ValueError("No Process")
        return self._process.signalProcess(signal_id)

    def __call__(self, hook):
        """
        Given the complete path name to a hook execute it with a
        runtime context.

        `hook`: The name of the hook to execute.
        """
        from twisted.internet import reactor

        # Sanity check the hook.
        self.validate_hook(hook)

        env = self.get_environment()
        hook_proto = HookProtocol(hook, self._context, self._log)
        deferred = hook_proto.deferred

        def cleanup_process(result):
            # Result should be the exit code of the process
            # clean up the connection and return the exit code
            # not the result of loseConnection.
            self._process.loseConnection()
            message = result
            if isinstance(message, Failure):
                message = message.getTraceback(elideFrameworkCode=True)
            self._log.debug("hook %s ended, exit code %s." % (
                    os.path.basename(hook), message))

            # Flush context changes back to zookeeper if hook was successful.
            if result == 0 and self._context:
                self._log.debug("Flushing relation settings changes ...")
                d = self._context.flush()
                d.addCallback(lambda _: result)
                return d

            return result

        self._process = reactor.spawnProcess(
            hook_proto, hook, [hook], env,
            os.path.join(self._unit_path, "formula"))
        deferred.addBoth(cleanup_process)

        return deferred
