import copy

from twisted.internet.defer import inlineCallbacks, returnValue, fail

from ensemble.errors import ProviderError
from ensemble.state.auth import make_identity

from .utils import get_user_authorized_keys

DEFAULT_REPOSITORIES = [
    "ppa:ensemble/ppa"]

DEFAULT_PACKAGES = [
    "bzr",
    "tmux",
    "python-setuptools",
    "python-twisted",
    "python-argparse",
    "python-txaws",
    "python-zookeeper"]

BOOTSTRAP_PACKAGES = [
    "bzr",
    "default-jre-headless",
    "zookeeper",
    "zookeeperd"]


def _get_initialize_script(instance_id_command, admin_secret):
    admin_identity = make_identity("admin:%s" % admin_secret)
    template = ('ensemble-admin initialize '
               '--instance-id=%s --admin-identity="%s"')
    return template % (instance_id_command, admin_identity)


def _get_machine_agent_script(vars):
    template = (
        "ENSEMBLE_MACHINE_ID=%(machine-id)s "
        "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
        "python -m ensemble.agents.machine -n "
        "--logfile=/var/log/ensemble/machine-agent.log "
        "--pidfile=/var/run/ensemble/machine-agent.pid")
    return template % vars


def _get_provision_agent_script(vars):
    template = (
        "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
        "python -m ensemble.agents.provision -n "
        "--logfile=/var/log/ensemble/provision-agent.log "
        "--pidfile=/var/run/ensemble/provision-agent.pid")
    return template % vars


class LaunchMachine(object):
    """Abstract class with generic instance-launching logic.

    Constructing with bootstrap=True will cause the run method to
    construct a machine which is also running a zookeeper for the
    cluster, and a provisioning agent, as well as the usual machine
    agent.

    To create your own subclass, you will certainly need to override
    C{start_machine} and C{get_instance_id_command}, and may find it
    convenient to override C{get_machine_variables} as well.
    """

    def __init__(self, provider, bootstrap=False):
        self._provider = provider
        self._bootstrap = bootstrap

    def run(self, machine_data):
        """Launch an instance node within the machine provider environment.

        @param machine_data a dictionary of data that is passed along to
               provided to the machine as serialized yaml. 'machine-id' is a
               required key, denoting the machine's zookeeper node for its
               machine agent.
        """
        if "machine-id" not in machine_data:
            return fail(ProviderError(
                "Machine state `machine-id` not provided in machine_data."))

        launch_deferred = self.get_machine_variables(machine_data)
        launch_deferred.addCallback(self.start_machine)
        if self._bootstrap:
            launch_deferred.addCallback(self._on_bootstrap_launched)
        return launch_deferred

    def start_machine(self, variables):
        """Actually launch a machine for the appropriate provider.

        @param variables: non-provider-specific data, sufficient to
            define the machine's behaviour once it exists.

        @return: a list of newly-launched ProviderMachines.
        """
        raise NotImplementedError()

    def get_instance_id_command(self):
        """Snippet to discover local instance-id

        @return: a snippet of shell script that evaluates to the
            local machine's instance_id.
        """
        raise NotImplementedError()

    @inlineCallbacks
    def get_machine_variables(self, vars_in):
        """Get the variables used for launching a machine and its cloud-init"""
        vars_out = copy.deepcopy(vars_in)
        vars_out[
            "ensemble-zookeeper-hosts"] = yield self._get_zookeeper_hosts()

        authorized_keys = get_user_authorized_keys(self._provider.config)
        vars_out["authorized_keys"] = authorized_keys
        vars_out["packages"] = self._get_packages()
        vars_out["repositories"] = list(DEFAULT_REPOSITORIES)
        vars_out["scripts"] = self._get_launch_scripts(vars_out)
        returnValue(vars_out)

    @inlineCallbacks
    def _get_zookeeper_hosts(self):
        if self._bootstrap:
            returnValue("localhost:2181")
        machines = yield self._provider.get_zookeeper_machines()
        hosts = [m.private_dns_name for m in machines]
        # format multiple hosts in string suitable for client
        returnValue(",".join(["%s:2181" % h for h in hosts]))

    def _get_packages(self):
        result = list(DEFAULT_PACKAGES)
        if self._bootstrap:
            result.extend(BOOTSTRAP_PACKAGES)
        return result

    def _get_launch_scripts(self, machine_data):
        """Get the scripts that will execute after the machine has launched."""
        launch_scripts = self._get_install_scripts()

        # These should be in the deb
        launch_scripts.extend([
            "sudo mkdir -p /var/lib/ensemble",
            "sudo mkdir -p /var/log/ensemble"])

        if self._bootstrap:
            launch_scripts.append(self._get_initialize_script())

        # Every machine has its own agent.
        machine_agent_script_template = (
            "ENSEMBLE_MACHINE_ID=%(machine-id)s "
            "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
            "python -m ensemble.agents.machine -n "
            "--logfile=/var/log/ensemble/machine-agent.log "
            "--pidfile=/var/run/ensemble/machine-agent.pid")
        launch_scripts.append(machine_agent_script_template % machine_data)

        if self._bootstrap:
            launch_scripts.append(_get_provision_agent_script(machine_data))

        return launch_scripts

    def _get_install_scripts(self):
        branch = self._provider.config.get("ensemble-branch")
        if branch is None:
            return ["sudo apt-get -y install ensemble"]
        # If we're using a custom branch, pass it via cloud-init to the agents,
        # so the entire cluster uses it.
        return [
            "sudo apt-get install -y python-txzookeeper",
            "sudo mkdir -p /usr/lib/ensemble",
            "cd /usr/lib/ensemble && sudo /usr/bin/bzr co %s ensemble" % branch,
            "cd /usr/lib/ensemble/ensemble && sudo python setup.py develop"]

    def _get_initialize_script(self):
        admin_secret = self._provider.config["admin-secret"]
        return _get_initialize_script(
            self.get_instance_id_command(), admin_secret)

    def _on_bootstrap_launched(self, machines):
        instance_ids = [m.instance_id for m in machines]
        d = self._provider.save_state({"zookeeper-instances": instance_ids})
        d.addCallback(lambda _: machines)
        return d
