import logging

from twisted.internet.defer import fail, succeed

from ensemble.errors import EnvironmentNotFound, ProviderError
from ensemble.lib.testing import TestCase
from ensemble.providers.common.bootstrap import Bootstrap
from ensemble.providers.common.launch import (
    BOOTSTRAP_PACKAGES, DEFAULT_PACKAGES, DEFAULT_REPOSITORIES, LaunchMachine)
from ensemble.providers.dummy import DummyMachine

DEFAULT_INSTALL = ["sudo apt-get -y install ensemble"]
CREATE_DIRS = ["sudo mkdir -p /var/lib/ensemble",
               "sudo mkdir -p /var/log/ensemble"]


class DummyLaunchMachine(LaunchMachine):

    def start_machine(self, variables):
        if self._bootstrap:
            name = "bootstrapped-instance-id"
        else:
            name = "some-instance-id"
        return [DummyMachine(name)]

    def get_instance_id_command(self):
        return "$(magic)"


def get_provider(launch_class=DummyLaunchMachine, config=None, zookeeper=True):
    class DummyProvider(object):
        config = {"authorized-keys": "abc"}
        launch_machine_class = launch_class

        def get_zookeeper_machines(self):
            if isinstance(zookeeper, Exception):
                return fail(zookeeper)
            return succeed(
               [DummyMachine("zookeeper-instance-id",
                             private_dns_name="zookeeper.internal")])

        def save_state(self, state):
            self.saved_state = state
            return succeed(None)

    if config is not None:
        DummyProvider.config.update(config)
    return DummyProvider()


def get_launch(launch_class=DummyLaunchMachine, config=None,
               zookeeper=True, bootstrap=False):
    provider = get_provider(launch_class, config, zookeeper)
    return launch_class(provider, bootstrap=bootstrap)


class LaunchMachineTest(TestCase):

    def _verify_machines(self, machines):
        (machine,) = machines
        self.assertTrue(isinstance(machine, DummyMachine))
        self.assertEquals(machine.instance_id, "some-instance-id")

    def test_no_machine_id(self):
        launch = get_launch()
        d = launch.run({})
        self.assertFailure(d, ProviderError)
        return d

    def test_basic_run(self):
        launch = get_launch()
        d = launch.run({"machine-id": "machine-77"})
        d.addCallback(self._verify_machines)
        return d

    def test_bootstrap_run(self):
        provider = get_provider(config={"admin-secret": "whatever"})
        launch = DummyLaunchMachine(provider, bootstrap=True)
        d = launch.run({"machine-id": "machine-32767"})

        def verify_state_saved(_):
            self.assertEquals(
                provider.saved_state,
                {"zookeeper-instances": ["bootstrapped-instance-id"]})
        d.addCallback(verify_state_saved)
        return d

    def test_get_machine_variables_normal(self):
        launch = get_launch()
        d = launch.get_machine_variables({"machine-id": "machine-999"})

        def verify_vars(vars):
            expect_vars = {
                "authorized_keys": "abc",
                "ensemble-zookeeper-hosts": "zookeeper.internal:2181",
                "machine-id": "machine-999",
                "packages": DEFAULT_PACKAGES,
                "repositories": DEFAULT_REPOSITORIES,
                "scripts": DEFAULT_INSTALL + CREATE_DIRS +
                           ["ENSEMBLE_MACHINE_ID=machine-999 "
                            "ENSEMBLE_ZOOKEEPER=zookeeper.internal:2181 "
                            "python -m ensemble.agents.machine -n "
                            "--logfile=/var/log/ensemble/machine-agent.log "
                            "--pidfile=/var/run/ensemble/machine-agent.pid"]}
            self.assertEquals(vars, expect_vars)
        d.addCallback(verify_vars)
        return d

    def test_get_machine_variables_with_ensemble_branch(self):
        branch_name = "lp:~my/ensemble/branch"
        launch = get_launch(config={"ensemble-branch": branch_name})
        d = launch.get_machine_variables({"machine-id": "machine-303"})

        def verify_vars(vars):
            scripts = vars["scripts"]
            install_scripts = ["sudo apt-get install -y python-txzookeeper",
                               "sudo mkdir -p /usr/lib/ensemble",
                               "cd /usr/lib/ensemble && "
                               "sudo /usr/bin/bzr co "
                               + branch_name + " ensemble",
                               "cd /usr/lib/ensemble/ensemble && "
                               "sudo python setup.py develop"]
            install_count = len(install_scripts)
            self.assertEquals(scripts[:install_count], install_scripts)

            other_scripts = scripts[install_count:]
            self.assertEquals(other_scripts,
                              CREATE_DIRS + [
                              "ENSEMBLE_MACHINE_ID=machine-303 "
                              "ENSEMBLE_ZOOKEEPER=zookeeper.internal:2181 "
                              "python -m ensemble.agents.machine -n "
                              "--logfile=/var/log/ensemble/machine-agent.log "
                              "--pidfile=/var/run/ensemble/machine-agent.pid"])
        d.addCallback(verify_vars)
        return d

    def test_get_machine_variables_bootstrap(self):
        launch = get_launch(config={"admin-secret": "SEEKRIT"},
                            bootstrap=True)
        d = launch.get_machine_variables({"machine-id": "machine-757"})

        def verify_vars(vars):
            expect_vars = {
                "authorized_keys": "abc",
                "ensemble-zookeeper-hosts": "localhost:2181",
                "machine-id": "machine-757",
                "packages": DEFAULT_PACKAGES + BOOTSTRAP_PACKAGES,
                "repositories": DEFAULT_REPOSITORIES}
            for key, value in expect_vars.items():
                self.assertEquals(vars[key], value)

            scripts = vars["scripts"]
            self.assertEquals(scripts[:3], DEFAULT_INSTALL + CREATE_DIRS)
            admin_init, machine_agent, provision_agent = scripts[3:]

            self.assertTrue(admin_init.startswith(
                "ensemble-admin initialize "
                '--instance-id=$(magic) --admin-identity="admin:'))
            self.assertFalse("SEEKRIT" in admin_init)

            expect_machine = (
                "ENSEMBLE_MACHINE_ID=machine-757 "
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.machine -n "
                "--logfile=/var/log/ensemble/machine-agent.log "
                "--pidfile=/var/run/ensemble/machine-agent.pid")
            self.assertEquals(machine_agent, expect_machine)

            expect_provision = (
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.provision -n "
                "--logfile=/var/log/ensemble/provision-agent.log "
                "--pidfile=/var/run/ensemble/provision-agent.pid")
            self.assertEquals(provision_agent, expect_provision)
        d.addCallback(verify_vars)
        return d


class BootstrapTest(TestCase):

    def test_bootstrap_unknown_error(self):
        class BBQError(Exception):
            pass
        provider = get_provider(zookeeper=BBQError())
        bootstrap = Bootstrap(provider)
        d = bootstrap.run()
        self.assertFailure(d, BBQError)
        return d

    def test_bootstrap_no_launch(self):
        log = self.capture_logging("ensemble.common", level=logging.DEBUG)
        provider = get_provider()
        bootstrap = Bootstrap(provider)
        d = bootstrap.run()

        def verify_machines(machines):
            (machine,) = machines
            self.assertTrue(isinstance(machine, DummyMachine))
            self.assertEquals(machine.instance_id, "zookeeper-instance-id")
            log_text = log.getvalue()
            self.assertIn(
                "Ensemble environment previously bootstrapped", log_text)
            self.assertNotIn("Launching", log_text)
        d.addCallback(verify_machines)
        return d

    def test_bootstrap_launch(self):
        log = self.capture_logging("ensemble.common", level=logging.DEBUG)
        provider = get_provider(config={"admin-secret": "SEEKRIT"},
                                zookeeper=EnvironmentNotFound())
        bootstrap = Bootstrap(provider)
        d = bootstrap.run()

        def verify_machines(machines):
            (machine,) = machines
            self.assertTrue(isinstance(machine, DummyMachine))
            self.assertEquals(machine.instance_id, "bootstrapped-instance-id")
            log_text = log.getvalue()
            self.assertIn("Launching Ensemble bootstrap instance", log_text)
            self.assertNotIn("previously bootstrapped", log_text)
        d.addCallback(verify_machines)
        return d
