import logging
import os

from yaml import safe_dump

from twisted.python.failure import Failure

from ensemble.errors import EnsembleError, ProviderInteractionError

log = logging.getLogger("ensemble.common")


def convert_unknown_error(failure):
    """Convert any non ensemble errors to a provider interaction error.

    Supports both usage from within an except clause, and as an
    errback handler ie. both the following forms are supported.

    ...
    try:
       something()
    except Exception, e:
       convert_unknown_errors(e)

    ...
    d.addErrback(convert_unknown_errors)
    """
    if isinstance(failure, Failure):
        error = failure.value
    else:
        error = failure

    if not isinstance(error, EnsembleError):
        error = ProviderInteractionError(error)

    if isinstance(failure, Failure):
        return Failure(error)
    raise error


def get_user_authorized_keys(config):
    """Locate a public key for the user.

    If neither "authorized-keys" nor "authorized-keys-path" is
    present in config, will look in the user's .ssh directory.

    @return: a public key.

    @raise LookupError: Raised if a public ssh key is not found.
    """
    key_names = ["id_dsa.pub", "id_rsa.pub", "identity.pub"]
    if config.get("authorized-keys"):
        return config["authorized-keys"]

    if config.get("authorized-keys-path"):
        key_names[:] = [config.get("authorized-keys-path")]

    for key_path in key_names:
        path_candidate = os.path.expandvars(os.path.expanduser(key_path))
        if not os.path.exists(path_candidate):
            path_candidate = "~/.ssh/%s" % key_path
            path_candidate = os.path.expanduser(path_candidate)

        if not os.path.exists(path_candidate):
            continue
        return open(path_candidate).read()
    raise LookupError("SSH authorized/public key not found.")


def format_cloud_ini(
    authorized_keys, packages=(), repositories=None, scripts=None, data=None):
    """Format a user-data cloud-init file.

    This will enable package installation, and ssh access, and script
    execution on launch, and passing data values to an instance.

    Its important to note that no sensistive data (credentials) should be
    conveyed via cloud-config, as the values are accessible from the any
    process by default on the instance.

    Further documentation on the capabilities of cloud-init
    https://help.ubuntu.com/community/CloudInit

    @param authorized_keys The authorized SSH keys to be used when populating
           the newly launched machine's.
    @type string

    @param packages A list of packages to be installed on a machine.
    @type list of strings

    @param repository A listing of debian repostiories to add the machine's
           apt sources. 'ppa:' syntax can be utilized.
    @type list of strings

    @param scripts A list of scripts to be executed on machine start.
    @type list of strings

    @param data An optional dictionary of data to be passed via cloud config.
           It will be accessible via the key 'machine-data' from the yaml
           data structure.
    """
    cloud_config = {
        "apt-update": True,
        "apt-upgrade": True,
        "ssh_authorized_keys": [authorized_keys],
        "packages": []}

    if data:
        cloud_config["machine-data"] = data

    if packages:
        cloud_config["packages"].extend(packages)

    if repositories:
        sources = [dict(source=r) for r in repositories]
        cloud_config["apt_sources"] = sources

    if scripts:
        cloud_config["runcmd"] = scripts

    output = safe_dump(cloud_config)
    output = "#cloud-config\n%s" % (output)
    return output
