import os
import copy

from txaws.service import AWSServiceRegion

from ensemble.environment.errors import EnvironmentsConfigError
from ensemble.providers.common.bootstrap import Bootstrap
from ensemble.providers.common.utils import (
    convert_unknown_error, get_user_authorized_keys)

from .accessor import EC2MachineAccessor
from .connect import EC2Connect
from .files import FileStorage
from .findzookeepers import EC2FindZookeepers
from .iterate import EC2MachineIteration
from .launch import EC2LaunchMachine
from .shutdown import EC2Shutdown, EC2ShutdownMachine
from .state import EC2SaveState, EC2LoadState
from .utils import get_region_uri


class MachineProvider(object):

    launch_machine_class = EC2LaunchMachine

    def __init__(self, environment_name, config):
        self.environment_name = environment_name
        self.config = config

        if not config.get("ec2-uri"):
            ec2_uri = get_region_uri(config.get("region", "us-east-1"))
        else:
            ec2_uri = config.get("ec2-uri")

        self._service = AWSServiceRegion(
            access_key=config.get("access-key", ""),
            secret_key=config.get("secret-key", ""),
            ec2_uri=ec2_uri,
            s3_uri=config.get("s3-uri", ""))
        self.s3 = self._service.get_s3_client()
        self.ec2 = self._service.get_ec2_client()

        if ("authorized-keys-path" in config and
            "authorized-keys" in config):
            raise EnvironmentsConfigError(
                "Environment config cannot define both authorized-keys "
                "and authorized-keys-path. Pick one!")

    def get_serialization_data(self):
        """Return a dictionary serialization of the provider configuration.

        Additionally this extracts crednetial information from the environment.
        """
        data = copy.deepcopy(self.config)
        data["secret-key"] = self.config.get(
            "secret-key", os.environ.get("AWS_SECRET_ACCESS_KEY"))
        data["access-key"] = self.config.get(
            "access-key", os.environ.get("AWS_ACCESS_KEY_ID"))
        data["authorized-keys"] = get_user_authorized_keys(data)

        # Not relevant, on a remote system.
        data.pop("authorized-keys-path", None)

        return data

    def _run_operation(self, operation, *args, **kw):
        d = operation.run(*args, **kw)
        d.addErrback(convert_unknown_error)
        return d

    def connect(self, share=False):
        """Connect to the zookeeper ensemble running in the machine provider.

        @param share: Requests sharing of the connection with other clients
            attempting to connect to the same provider, if that's feasible.

        returns an open C{txzookeeper.client.ZookeeperClient} and a
        C{ensemble.storage.connection.TunnelProtocol}
        """
        connect = EC2Connect(self)
        return self._run_operation(connect, share=share)

    def get_file_storage(self):
        """Retrieve the provider C{FileStorage} abstraction."""
        file_storage = FileStorage(self.s3, self.config["control-bucket"])
        return file_storage

    def list_machines(self):
        """List all machines running in the provider."""
        iteration = EC2MachineIteration(self)
        return iteration.run()

    def get_machine(self, provider_machine_id):
        """Retrieve a provider machine by provider machine id."""
        machine_accessor = EC2MachineAccessor(self)
        return machine_accessor.run(provider_machine_id)

    def start_machine(self, machine_data):
        """Start a machine in the provider.

        @param machine_data a dictionary of data to pass along to the newly
               launched machine.
        @type dict
        """
        launch = EC2LaunchMachine(self)
        return launch.run(machine_data=machine_data)

    def shutdown_machine(self, machine):
        """Stop a machine in the provider."""
        shutdown = EC2ShutdownMachine(self)
        return shutdown.run(machine)

    def bootstrap(self):
        """Bootstrap an ensemble server in the provider."""
        bootstrap = Bootstrap(self)
        return bootstrap.run()

    def shutdown(self):
        """Shutdown all machine resources associated to this provider instance.
        """
        shutdown = EC2Shutdown(self)
        return shutdown.run()

    def save_state(self, state):
        """Save state to the provider.

        @param state
        @type dict
        """
        serializer = EC2SaveState(self)
        return serializer.run(state)

    def load_state(self):
        """Load state from the provider.

        @return: a dictionary.
        """
        deserializer = EC2LoadState(self)
        return deserializer.run()

    def get_zookeeper_machines(self):
        """Find running zookeeper instances.

        @return: the first valid instance found as a single element list.

        @raise: EnvironmentNotFound or EnvironmentPending
        """
        find_zookeepers = EC2FindZookeepers(self)
        return find_zookeepers.run()
