from twisted.internet.defer import succeed

from txaws.ec2.exception import EC2Error

from ensemble.errors import EnvironmentNotFound, EnvironmentPending

from .machine import machine_from_instance


class EC2FindZookeepers(object):

    def __init__(self, provider):
        self._provider = provider

    def run(self):
        """Verify which zookeeper nodes are running.

        Returns the first valid instance found as a single element list,
        or errors out.
        """
        state_deferred = self._provider.load_state()
        state_deferred.addCallback(self._check_state)
        return state_deferred

    def _check_state(self, state):
        instance_ids = None

        if state:
            instance_ids = state.get("zookeeper-instances", None)

        if not instance_ids:
            d = succeed([])
        else:
            instance_id = instance_ids.pop(0)
            d = self._provider.ec2.describe_instances(instance_id)

        d.addCallback(self._cb_filter_running_zookeeper_instances)
        d.addErrback(self._eb_invalid_instance_id, instance_ids)
        return d

    def _eb_invalid_instance_id(self, failure, instance_ids):
        # Trap these failures to continue processing instances
        failure.trap(
            EC2Error, EnvironmentPending, EnvironmentNotFound)

        # Reformat invalid ids ec2 errors into an appros ensemble error.
        if not instance_ids and failure.check(EC2Error):
            error_code = failure.value.get_error_codes()
            if error_code == "InvalidInstanceID.NotFound":
                raise EnvironmentNotFound(
                    "No Ensemble machines found. "
                    "Is the environment bootstrapped?")

        # Process next instance
        if instance_ids:
            instance_id = instance_ids.pop(0)
            d = self._provider.ec2.describe_instances(instance_id)

            d.addCallback(self._cb_filter_running_zookeeper_instances)
            d.addErrback(self._eb_invalid_instance_id, instance_ids)
            return d

        # Return the original failure
        return failure

    def _cb_filter_running_zookeeper_instances(self, instances):
        # instances is either a single element list or an empty list.
        assert len(instances) <= 1, "Should only process single instance %r"
        for instance in instances:
            if instance.instance_state == "running":
                return [machine_from_instance(instance)]
            elif instance.instance_state == "pending":
                raise EnvironmentPending(
                    "Started machine is still pending.")
        raise EnvironmentNotFound(
            "No Ensemble machines found. "
            "Is the environment bootstrapped?")
