from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.providers.common.launch import LaunchMachine

from .machine import machine_from_instance
from .utils import log, get_launch_options


class EC2LaunchMachine(LaunchMachine):
    """Amazon EC2 operation for launching an instance"""

    @inlineCallbacks
    def start_machine(self, variables):
        """Actually launch an instance on EC2.

        @param variables: should be a dictionary with entries sufficient
            to specify everything necessary to launch the requested
            instance.

        @return: a singe-entry list containing a ProviderMachine for the
            new instance
        """
        # Retrieves standard ec2 run_instances arguments, of note
        # image id, and instance type.
        machine_options = yield get_launch_options(
            self._provider.config,
            **variables)

        # Ensure the ensemble security group exists and is included in the
        # machine instance launch options.
        machine_options["security_groups"] = yield self._ensure_group(
            machine_options["security_groups"])

        # Launch the instance.
        instances = yield self._provider.ec2.run_instances(**machine_options)
        returnValue([machine_from_instance(instance)
                     for instance in instances])

    def get_instance_id_command(self):
        """Snippet to discover local instance-id

        @return: a snippet of shell script that evaluates to the
            local machine's instance_id.
        """
        return "$(curl http://169.254.169.254/1.0/meta-data/instance-id)"

    def get_machine_variables(self, vars_in):
        vars_deferred = super(
            EC2LaunchMachine, self).get_machine_variables(vars_in)

        def set_provider_type(vars):
            vars["ensemble-provider-type"] = "ec2"
            return vars
        vars_deferred.addCallback(set_provider_type)
        return vars_deferred

    def _ensure_group(self, groups):
        """Ensure the ensemble group is the machine launch groups.

        Machines launched by ensemble are tagged with a group so they can
        be distinguished from other machines that might be running on
        an EC2 account. This group can be specified explicitly or implicitly
        defined by the environment name.

        @param security_groups: The configured security groups for a machine
        instance.
        """
        # Label the instance with the ensemble name.
        d = self._provider.ec2.describe_security_groups()
        d.addCallback(self._on_received_groups, groups)
        return d

    def _on_received_groups(self, security_groups, groups):
        ensemble_group = "ensemble-%s" % self._provider.environment_name
        group_ids = [group.name for group in security_groups]

        # Ensure the provider group is added to the instance groups.
        if not ensemble_group in groups:
            groups.append(ensemble_group)

        # Create the provider group if doesn't exist.
        if not ensemble_group in group_ids:
            log.debug("Creating ensemble provider group %s", ensemble_group)
            group_created = self._provider.ec2.create_security_group(
                ensemble_group,
                "Ensemble group for %s" % self._provider.environment_name)

            # Authorize SSH.
            group_created.addCallback(
                lambda x: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="22", to_port="22",
                    cidr_ip="0.0.0.0/0"))

            # We need to describe the group to pickup the owner_id for auth.
            group_created.addCallback(
                lambda x: self._provider.ec2.describe_security_groups(
                    ensemble_group))

            # Authorize Internal ZK Traffic
            group_created.addCallback(
                lambda groups: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    source_group_name=ensemble_group,
                    source_group_owner_id=groups.pop().owner_id))

            # TODO: For now authorize external traffic to the instance,
            # except for the ZooKeeper port.  The expose functionality
            # that is in progress will address this in a better way.
            group_created.addCallback(
                lambda groups: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="1",
                    to_port="2180",
                    cidr_ip="0.0.0.0/0"))
            group_created.addCallback(
                lambda groups: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="2182",
                    to_port="65535",
                    cidr_ip="0.0.0.0/0"))

            group_created.addCallback(
                lambda groups: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="udp",
                    from_port="1",
                    to_port="2180",
                    cidr_ip="0.0.0.0/0"))
            group_created.addCallback(
                lambda groups: self._provider.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="udp",
                    from_port="2182",
                    to_port="65535",
                    cidr_ip="0.0.0.0/0"))

            # We use a separate callback here, because we need to wait
            # on the authorize security group deferred to succeed before
            # we can return the groups result.
            return group_created.addCallback(lambda x: groups)
        return groups
