from twisted.internet.defer import fail

from ensemble.errors import ProviderError

from .machine import EC2ProviderMachine, machine_from_instance
from .iterate import EC2ProviderMachineFilter


class EC2Shutdown(EC2ProviderMachineFilter):
    """Amazon EC2 MachineProvider environment shutdown."""

    def run(self):
        """
        Shutdown any EC2 machine instances associated to this
        machine provider instance.
        """
        d = self._provider.ec2.describe_instances()
        d.addCallback(self._filter_provider_machines)
        d.addCallback(self._instance_shutdown)
        return d

    def _instance_shutdown(self, instances):
        """Terminate all provided instances."""
        if not instances:
            return

        def on_shutdown(result):
            return [machine_from_instance(instance)
                    for instance in instances]

        deferred_shutdown = self._provider.ec2.terminate_instances(
            *[i.instance_id for i in instances])
        deferred_shutdown.addCallback(on_shutdown)
        return deferred_shutdown


class EC2ShutdownMachine(EC2Shutdown):
    """Amazon EC2 MachineProvider environment shutdown."""

    def run(self, machine):
        """
        Shutdown any EC2 machine instances in the given environment
        with the given provider name.
        """
        if not isinstance(machine, EC2ProviderMachine):
            return fail(ProviderError(
                "EC2 got a non-EC2 ProviderMachine (%r)" %
                type(machine)))
        instance_id = machine.instance_id
        d = self._provider.ec2.describe_instances(instance_id)
        d.addCallback(self._filter_provider_machines, True)
        d.addCallback(self._instance_shutdown)
        return d
