import base64
import random
import string
import yaml

from zookeeper import (
    BadArgumentsException, BadVersionException,
    NoNodeException, NodeExistsException, SEQUENCE)

from twisted.internet.defer import inlineCallbacks, returnValue
from txzookeeper.client import ZOO_OPEN_ACL_UNSAFE

from ensemble.state.auth import make_identity, make_ace
from ensemble.state.errors import (
    StateChanged, StateNotFound, PrincipalNotFound)
from ensemble.state.utils import YAMLState
from ensemble.state import securityrules

ZOO_OPEN_AUTH_ACL_UNSAFE = ZOO_OPEN_ACL_UNSAFE
OTP_IDENTITY_KEY = "otp-identity"


def _generate_string(size=16):
    return "".join(random.sample(string.letters, size))


@inlineCallbacks
def apply_rules(client, path):
    """Apply security policy ACL to the given path."""
    token_db = TokenDatabase(client)
    policy = SecurityPolicy(
        client, token_db, securityrules.get_default_rules())
    acl = yield policy(path)
    yield client.set_acl(path, acl)


def get_domain_security(client, path):
    """Get a unified security convience api that operates against a node.
    """
    return _DomainSequenceSecurity(client, path)


def set_domain_security(enabled):
    """For testing enable disabling security for test speed.

    All the operations against DomainSequenceSecurity become no-ops.
    """
    _DomainSequenceSecurity.enabled = bool(enabled)


class _DomainSequenceSecurity(object):
    """Sugar syntax around typical sequence domain object usage.

    When a domain object representing a connected client, the typical
    tasks that need to be done are quite similiar. This class embodies
    the work of creating and storing an OTP serilaization to a node,
    and applying security rules using a supplied topology to the
    domain node.
    """

    enabled = True

    def __init__(self, client, path):
        self._client = client
        self._path = path

    @inlineCallbacks
    def apply_otp(self, principal_name):
        """Create an otp principal and store its serialization at path.

        Also stores the named principal in the token database, so it
        may used immediately for ACL construction. Returns the principal.
        """
        if not self.enabled:
            return

        principal = OTPPrincipal(self._client)
        yield principal.create(principal_name)
        token_db = TokenDatabase(self._client)
        yield token_db.add(principal)

        state = YAMLState(self._client, self._path)
        yield state.read(required=True)

        if OTP_IDENTITY_KEY in state:
            raise RuntimeError("OTP Identity already exists")
        else:
            state[OTP_IDENTITY_KEY] = principal.serialize()
        yield state.write()
        returnValue(principal)

    @inlineCallbacks
    def apply_rules(self, topology):
        """Apply security rules to the node.

        TODO: Topology support for policy application.
        """
        if not self.enabled:
            return
        yield apply_rules(self._client, self._path)

    @inlineCallbacks
    def consume_otp(self):
        """Consumes an OTP serialization and returns the permanent principal.

        The OTP data is used to inform agents how they can retrieve
        their permanent credentials. This function will consume that
        data retrieve persistent prinicipal, and return it.
        """
        if not self.enabled:
            return

        try:
            content, stat = yield self._client.get(self._path)
        except NoNodeException:
            raise StateNotFound(self._path)

        data = yaml.load(content)

        if not data or not OTP_IDENTITY_KEY in data:
            raise PrincipalNotFound("OTP")

        credentials = yield OTPPrincipal.consume(
            self._client, data[OTP_IDENTITY_KEY])

        returnValue(Principal(*credentials))

    @inlineCallbacks
    def add_group(self, principal_name):
        """Create a group principal."""

        if not self.enabled:
            return

        token_db = TokenDatabase(self._client)
        principal = GroupPrincipal(self._client, "%s/group" % self._path)
        yield principal.create(principal_name, _generate_string())
        yield token_db.add(principal)
        returnValue(principal)

    def add_group_member(self, principal):
        """Add the principal to the group."""
        if not self.enabled:
            return
        group = GroupPrincipal(self._client, "%s/group" % self._path)
        return group.add_member(principal)

    def remove_group_member(self, principal_name):
        """Remove a principal by name from the group."""
        if not self.enabled:
            return
        group = GroupPrincipal(self._client, "%s/group" % self._path)
        return group.remove_member(principal_name)


class Principal(object):
    """An ensemble/zookeeper principal.
    """

    def __init__(self, name, password):
        self._name = name
        self._password = password

    @property
    def name(self):
        """A principal has a login name."""
        return self._name

    def get_token(self):
        """A principal identity token can be retrieved.

        An identity token is used to construct ACLs.
        """
        return make_identity("%s:%s" % (self._name, self._password))

    def attach(self, connection):
        """A principal can be attached to a connection."""
        auth_deferred = connection.add_auth(
            "digest", "%s:%s" % (self._name, self._password))
        # Work around for fast auth, remove post ZOOKEEPER-770
        connection.exists("/")
        return auth_deferred


class GroupPrincipal(object):
    """A group principal is a principal that can have multiple members.

    Group principals are persistent with their credentials stored in zk.
    Membership of the group allows a group member to retrieve and utilize
    these credentials.
    """

    def __init__(self, client, path, name=None):
        self._client = client
        self._path = path
        self._name = name
        self._password = None

    def _check(self):
        if self._name is None:
            raise ValueError("initialize() must be called before usage")

    @property
    def name(self):
        """A principal has a login name."""
        self._check()
        return self._name

    def get_token(self):
        """A principal identity token can be retrieved.

        An identity token is used to construct ACLs.
        """
        self._check()
        return make_identity("%s:%s" % (self._name, self._password))

    @inlineCallbacks
    def attach(self, connection):
        self._check()
        auth_deferred = connection.add_auth(
            "digest", "%s:%s" % (self._name, self._password))
        # Work around for fast auth, remove post ZOOKEEPER-770
        connection.exists("/")
        yield auth_deferred

    @inlineCallbacks
    def create(self, name, password):
        """Create the group with the given name and password."""
        try:
            yield self._client.create(
                self._path,
                yaml.safe_dump(dict(name=name, password=password)))
        except NodeExistsException:
            raise ValueError("Group already exists at %r" % self._path)
        self._name = name
        self._password = password

    @inlineCallbacks
    def initialize(self):
        """Initialize the group.

        The group must always be initialized before attribute
        use. Groups are persistent principals. If group is being used
        for the first time it must be initialized with its name and
        password.
        """

        try:
            data, stat = yield self._client.get(self._path)
        except NoNodeException:
            raise StateNotFound("group does not exist at %r" % self._path)

        credentials = yaml.load(data)
        self._name = credentials["name"]
        self._password = credentials["password"]
        returnValue(self)

    def add_member(self, principal):
        """Add a principal as a member of the group.

        A member of the group can use the group as an additional principal
        attached to the connection.
        """
        acl = ACL(self._client, self._path)
        return acl.grant_token(principal.get_token(), read=True)

    def remove_member(self, name):
        """Remove a principal from a group by principal name"""
        acl = ACL(self._client, self._path)
        return acl.prohibit(name)

    @inlineCallbacks
    def get_members(self):
        """Return the names of the principals that are members of the group."""
        acl, stat = yield self._client.get_acl(self._path)
        results = []
        for entry in acl:
            results.append(entry["id"].split(":", 1)[0])
        returnValue(results)


class OTPPrincipal(object):
    """One Time Password (OTP) Principal.

    Its common for ensemble to need to pass credentials for newly
    created principals over unsecure channels to external
    processes. In order to mitigate the risks of interception of these
    credentials, a one time password is used that enables the
    retrieval of the intended principal credential by the external
    process.
    """

    # Additional OTP node ACL entry, see :method set_otp_additional_ace:
    _extra_otp_ace = None

    def __init__(self, client, path="/otp/otp-"):
        self._client = client
        self._name = None
        self._password = None
        self._otp_name = None
        self._otp = None
        self._path = path

    def _check(self):
        if not self._name:
            raise ValueError("OTPPrincipal must be created before use")

    @property
    def name(self):
        """A principal has a login name."""
        self._check()
        return self._name

    def get_token(self):
        """A principal identity token can be retrieved.

        An identity token is used to construct ACLs.
        """
        self._check()
        return make_identity("%s:%s" % (self._name, self._password))

    @inlineCallbacks
    def attach(self, connection):
        raise NotImplemented("OTP Principals shouldn't attach")

    @classmethod
    @inlineCallbacks
    def consume(cls, client, otp_data):
        """Consume an otp serialization to retrieve the actual credentials.

        returns a username, password tuple, and destroys the otp node.
        """
        # Decode the data to get the path and otp credentials
        path, credentials = base64.b64decode(otp_data).split(":", 1)
        attach_deferred = client.add_auth("digest", credentials)
        # Fast Auth - Remove after ZOOKEEPER-770
        client.exists("/")

        yield attach_deferred
        # Load the otp principal data
        data, stat = yield client.get(path)
        principal_data = yaml.load(data)
        # Consume the otp node
        yield client.delete(path)
        returnValue((principal_data["name"], principal_data["password"]))

    @inlineCallbacks
    def create(self, name, password=None, otp_name=None, otp=None):
        """Create an OTP for a principal.
        """
        if self._name:
            raise ValueError("OTPPrincipal has already been created.")

        self._name = name
        self._password = password or _generate_string()
        self._otp_name = otp_name or _generate_string()
        self._otp = otp or _generate_string()

        acl = [make_ace(
            make_identity("%s:%s" % (self._otp_name, self._otp)), read=True)]

        # Optional additional ACL entry for unit test teardown.
        if self._extra_otp_ace:
            acl.append(self._extra_otp_ace)

        self._path = yield self._client.create(
            self._path,
            yaml.safe_dump(dict(name=self._name, password=self._password)),
            acls=acl,
            flags=SEQUENCE)

        returnValue(self)

    def serialize(self):
        """Return a serialization of the otp path and credentials.

        This can be sent to an external process such that they can access
        the otp identity.
        """
        return base64.b64encode(
            "%s:%s:%s" % (self._path, self._otp_name, self._otp))

    @classmethod
    def set_additional_otp_ace(cls, ace):
        """This method sets an additional ACl entry to be added to OTP nodes.

        This method is meant for testing only, to ease construction of unit
        test tear downs when OTP nodes are created.
        """
        cls._extra_otp_ace = ace


class TokenDatabase(object):
    """A hash map of principal names to their identity tokens.

    Identity tokens are used to construct node ACLs.
    """
    def __init__(self, client, path="/auth-tokens"):
        self._state = YAMLState(client, path)

    @inlineCallbacks
    def add(self, principal):
        """Add a principal to the token database.
        """
        yield self._state.read()
        self._state[principal.name] = principal.get_token()
        yield self._state.write()

    @inlineCallbacks
    def get(self, name):
        """Return the identity token for a principal name.
        """
        yield self._state.read()
        try:
            returnValue(self._state[name])
        except KeyError:
            raise PrincipalNotFound(name)

    @inlineCallbacks
    def remove(self, name):
        """Remove a principal by name from the token database.
        """
        yield self._state.read()
        if name in self._state:
            del self._state[name]
            yield self._state.write()


class SecurityPolicy(object):
    """The security policy generates ACLs for new nodes based on their path.
    """
    def __init__(self, client, token_db, rules=(), owner=None):
        self._client = client
        self._rules = list(rules)
        self._token_db = token_db
        self._owner = None

    @property
    def client(self):
        return self._client

    def set_owner(self, principal):
        """If an owner is set all nodes ACLs will grant access to the owner.
        """
        assert not self._owner, "Owner already assigned"
        self._owner = principal

    def add_rule(self, rule):
        """Add a security rule to the policy.

        A rule is a callable object accepting the policy and the path as
        arguments. The rule should return a list of ACL entries that apply
        to the node at the given path. Rules may return deferred values.
        """
        self._rules.append(rule)

    def get_token(self, principal_name):
        return self._token_db.get(principal_name)

    @inlineCallbacks
    def __call__(self, path):
        """Given a node path, determine the ACL.
        """
        acl_entries = []

        for rule in self._rules:
            entries = yield rule(self, path)
            if entries:
                acl_entries.extend(entries)

        # XXX/TODO - Remove post security-integration
        # Allow incremental integration
        if not acl_entries:
            acl_entries.append(ZOO_OPEN_AUTH_ACL_UNSAFE)

        # Give cli admin access by default
        admin_token = yield self._token_db.get("admin")
        acl_entries.append(make_ace(admin_token, all=True))

        # Give owner access by default
        if self._owner:
            acl_entries.append(
                make_ace(self._owner.get_token(), all=True))

        returnValue(acl_entries)


class SecureConnectionMixin(object):

    _policy = None

    def set_security_policy(self, policy):
        self._policy = policy

    @inlineCallbacks
    def create(self, path, data="", acls=[ZOO_OPEN_ACL_UNSAFE], flags=0):
        """Creates a zookeeper node at the given path, with the given data.

        The secure connection mixin, defers acl values to a security policy
        set on the connection if any.
        """
        if self._policy and acls == [ZOO_OPEN_ACL_UNSAFE]:
            acls = yield self._policy(path)
        result = yield super(SecureConnectionMixin, self).create(
            path, data, acls=acls, flags=flags)
        returnValue(result)


class ACL(object):
    """A ZooKeeper Node ACL.

    Allows for permission grants and removals to principals by name.
    """

    def __init__(self, client, path):
        self._client = client
        self._path = path
        self._token_db = TokenDatabase(client)

    @inlineCallbacks
    def grant(self, principal_name, **perms):
        """Grant permissions on node to the given principal name."""

        token = yield self._token_db.get(principal_name)
        yield self.grant_token(token, **perms)

    @inlineCallbacks
    def grant_token(self, token, **perms):
        """Grant permissions on node to the given identity token.
         """
        ace = make_ace(token, **perms)
        principal_name = token.split(":")[0]

        def add(acl):
            if self._principal_index(acl, principal_name) is not None:
                raise StateChanged(
                    "Principal %r already part of ACL @ %r" % (
                        principal_name, self._path))
            acl.append(ace)
            return acl

        yield self._update_acl(add)

    @inlineCallbacks
    def prohibit(self, principal_name):
        """Remove all grant for the given principal name."""

        def remove(acl):
            index = self._principal_index(acl, principal_name)
            if index is None:
                # We got to the same end goal.
                return acl
            acl.pop(index)
            return acl

        yield self._update_acl(remove)

    @inlineCallbacks
    def _update_acl(self, change_func):
        """Update an ACL using the given change function to get a new acl.

        Goal is to be tolerant of non-conflicting concurrent updates.
        """
        while True:
            try:
                acl, stat = yield self._client.get_acl(self._path)
            except (BadArgumentsException, NoNodeException):
                raise StateNotFound(self._path)

            acl = change_func(acl)
            try:
                yield self._client.set_acl(
                    self._path, acl, version=stat["aversion"])
            except BadVersionException:
                continue
            else:
                break

    def _principal_index(self, acl, principal_name):
        """Determine the index into the ACL of a given principal ACE."""
        for index in range(len(acl)):
            ace = acl[index]
            if ace["id"].split(":", 1)[0] == principal_name:
                return index
        return None
