"""
Security rules to determine ACL by node path.

"""

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.state.auth import make_ace
from ensemble.state.topology import InternalTopology


def formula_rule(policy, path):
    """An ACL policy rule for nodes under the '/formulas' path.

    These nodes are read only by any connected user, and are
    writable only by the admin cli (which creates them).

    Path rule::

      /formulas
         /formula-id (all:admin, r:everyone)
    """
    if not path.startswith("/formulas"):
        return
    return [make_ace("anyone", "world", read=True)]


@inlineCallbacks
def machine_rule(policy, path):
    """An ACL policy rule for nodes under the '/machines' path.

    These nodes are created by the cli admin when launching new machines,
    with admin access granted to the provisioning agents. When the
    provisioning agents create a new machine, they also create a new
    principal and update the machine node ace to allow access to the
    new machine agent principal.
    """

    if not path.startswith("/machines"):
        return

    agent_token = yield policy.get_token("ensemble-provider")

    # If its a machine state
    if path.startswith("/machines/machine-"):
        # The provisioning agent updates the acl with the machine agent.
        returnValue([make_ace(agent_token, read=True, admin=True)])
    # If its machine state container
    elif path == "/machines":
        # Its readable by the provisioning agent
        returnValue([make_ace(agent_token, read=True)])


@inlineCallbacks
def _get_service_data(policy, relation_id):
    """Discover service information for a relation

    :param policy: A :class:SecurityPolicy instance
    :param relation_id: The internal relation id

    Returns a list of dictionaries containing service relation data
    and a the service token.
    """
    topology_data, stat = yield policy.client.get("/topology")
    topology = InternalTopology()
    topology.parse(topology_data)
    services = topology.get_relation_services(relation_id)
    results = []
    for service_id, service_data in services.items():
        service_data = dict(service_data)
        service_data["service_id"] = service_id
        service_name = topology.get_service_name(service_id)
        service_data["token"] = yield policy.get_token(service_name)
        results.append(service_data)
    returnValue(results)


@inlineCallbacks
def relation_rule(policy, path):
    """An ACL policy rule for nodes under the '/relations' path.
    """
    if not path.startswith("/relations"):
        return

    if path == "/relations":
        return

    parts = filter(None, path.split("/"))[1:]

    # Extract relation-id
    relation_id = parts.pop(0)

    # Retrieve data for services in the relation
    service_data = yield _get_service_data(policy, relation_id)
    service_tokens = [data["token"] for data in service_data]

    # The relation itself /relations/relation-id
    if not parts:
        returnValue(map(lambda x: make_ace(x, read=True), service_tokens))

    container = parts.pop(0)

    # If its just the container /relations/relation-id/(settings|role)
    # then allow both services to read from it
    if not parts:
        if container in ("settings", "peer"):
            returnValue(
                map(lambda x: make_ace(x, create=True, read=True),
                    service_tokens))
        elif container in ("client", "server"):
            matched_service = [data for data in service_data \
                               if data["role"] == container].pop()
            related_service = [data for data in service_data \
                               if data != matched_service].pop()
            returnValue(
                [make_ace(matched_service["token"], read=True, create=True),
                 make_ace(related_service["token"], read=True)])
    else:
        # Allow read access to settings and presence nodes
        returnValue(
            map(lambda x: make_ace(x, read=True), service_tokens))


@inlineCallbacks
def service_rule(policy, path):
    """An ACL rule for the /services hierarchy."""

    if not path.startswith("/services"):
        return

    if path == "/services":
        return

    parts = filter(None, path.split("/"))[1:]
    service_id = parts.pop(0)

    if not parts:
        return

    if parts[0] == "config":
        topology_data, stat = yield policy.client.get("/topology")
        topology = InternalTopology()
        topology.parse(topology_data)
        service_token = yield policy.get_token(
            topology.get_service_name(service_id))
        returnValue([make_ace(service_token, read=True)])
    elif parts[0] == "exposed":
        provider_token = yield policy.get_token("ensemble-provider")
        returnValue([make_ace(provider_token, read=True)])


@inlineCallbacks
def unit_rule(policy, path):
    """An ACL generator rule for the '/units' subtree."""
    if not path.startswith("/units"):
        return

    if path == "/units":
        return

    parts = filter(None, path.split("/"))[1:]
    unit_id = parts.pop(0)

    topology_data, stat = yield policy.client.get("/topology")
    topology = InternalTopology()
    topology.parse(topology_data)

    unit_name = topology.get_service_unit_name_from_id(unit_id)
    unit_token = yield policy.get_token(unit_name)

    if not parts:
        # The unit agent needs to write to the node to update the formula post
        # upgrade. it needs to delete children from the node when clearing
        # various flags. It needs to create when establishing open ports, or
        # its agent presence.
        returnValue([make_ace(
            unit_token, read=True, write=True, delete=True, create=True)])

    segment = parts.pop(0)

    if segment == "ports":
        # Created by the unit agent, read by the provisioning agent.
        provider_token = yield policy.get_token("ensemble-provider")
        returnValue([make_ace(provider_token, read=True)])
    elif segment in ("upgrade", "debug", "resolved", "relations-resolved"):
        # Created by the admin, read and deleted by the unit agent.
        returnValue([make_ace(unit_token, read=True)])


@inlineCallbacks
def global_rule(policy, path):
    """An ACL for top-level nodes like '/topology' and '/environment'
    """
    if path == "/environment":
        agent_token = yield policy.token_db.get_token("ensemble-provider")
        # Only the admin cli can write to this.
        returnValue([make_ace(agent_token, read=True)])
    elif path == "/topology":
        # Only the admin cli can write to this.
        returnValue([make_ace("auth", "world", read=True)])
    elif path == "/auth-tokens":
        # Any connection that creates another user needs to modify this
        returnValue([make_ace("auth", "world", read=True, write=True)])


DEFAULT_RULES = [
    global_rule,
    formula_rule,
    relation_rule,
    machine_rule,
    service_rule,
    unit_rule
    ]


def get_default_rules():
    return []
