from twisted.internet.defer import inlineCallbacks, returnValue

import zookeeper

from txzookeeper import ZookeeperClient
from txzookeeper.tests.utils import deleteTree

from ensemble.lib.testing import TestCase
from ensemble.formula.tests.test_directory import sample_directory
from ensemble.formula.directory import FormulaDirectory
from ensemble.state.auth import make_ace

from ensemble.state.security import (
    Principal, OTPPrincipal, TokenDatabase, set_domain_security)

from ensemble.state.topology import InternalTopology
from ensemble.tests.common import get_test_zookeeper_address


class StateTestBase(TestCase):

    @inlineCallbacks
    def setUp(self):
        yield super(StateTestBase, self).setUp()
        zookeeper.set_debug_level(0)
        self.formula = FormulaDirectory(sample_directory)
        self.client = self.get_zookeeper_client()

        yield self.client.connect()
        yield self.client.create("/formulas")
        yield self.client.create("/machines")
        yield self.client.create("/services")
        yield self.client.create("/units")
        yield self.client.create("/relations")
        yield self.client.create("/otp")

        ## Default this to on for now, about 20-40% increase in test time
        ## across individual state tests modules
        #yield self.setup_security()

        # Can be set to off except for explicitly marked tests via
        # "enable_security" decorator or setup_security used by the
        # test
        self._security_enabled = False
        set_domain_security(False)

    @property
    def security_enabled(self):
        return self._security_enabled

    @inlineCallbacks
    def tearDown(self):
        # Close and reopen connection, so that watches set during
        # testing are not affected by the cleaning up.
        self.client.close()
        client = ZookeeperClient(get_test_zookeeper_address())
        yield client.connect()

        if self._security_enabled:
            yield Principal("testadmin", "testadmin").attach(client)
            yield client.exists("/")

        deleteTree(handle=client.handle)
        client.close()
        yield super(StateTestBase, self).tearDown()

    @inlineCallbacks
    def get_topology(self):
        """Read /topology and return InternalTopology instance with it."""
        content, stat = yield self.client.get("/topology")
        topology = InternalTopology()
        topology.parse(content)
        returnValue(topology)

    @inlineCallbacks
    def set_topology(self, topology):
        """Dump the given InternalTopology into /topology."""
        content = topology.dump()
        try:
            yield self.client.set("/topology", content)
        except zookeeper.NoNodeException:
            yield self.client.create("/topology", content)

    @inlineCallbacks
    def setup_security(self):
        """Additional test setup to allow for OTP and security rules.
        """
        # Setup additional test access to OTP nodes
        principal = Principal("testadmin", "testadmin")
        token = principal.get_token()
        OTPPrincipal.set_additional_otp_ace(make_ace(token, all=True))
        # Rules application expects 'admin' principal
        yield TokenDatabase(self.client).add(Principal("admin", "admin"))
        # Remove the test otp ace, post test
        self.addCleanup(lambda: OTPPrincipal.set_additional_otp_ace(None))
        self._security_enabled = True
        set_domain_security(True)


def security_test(test_method):

    @inlineCallbacks
    def run_security_enabled_test(self, *args, **kw):
        if not self.security_enabled:
            yield self.setup_security()
        yield test_method(self, *args, **kw)

    return run_security_enabled_test
