import yaml

from twisted.internet.defer import inlineCallbacks

from ensemble.state.formula import FormulaStateManager
from ensemble.state.errors import FormulaStateNotFound

from ensemble.state.tests.common import StateTestBase


class FormulaStateManagerTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(FormulaStateManagerTest, self).setUp()
        self.formula_state_manager = FormulaStateManager(self.client)

    @inlineCallbacks
    def test_add_formula(self):
        """
        Adding a Formula into a FormulaStateManager should register
        the formula within the Zookeeper state, according to the
        specification.
        """
        formula_id = "namespace:dummy-1"

        formula_state = yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        self.assertEquals(formula_state.id, formula_id)

        children = yield self.client.get_children("/formulas")
        self.assertEquals(children, [formula_id])

        content, stat = yield self.client.get("/formulas/" + formula_id)

        formula_data = yaml.load(content)
        self.assertEquals(formula_data, {
            "metadata": self.formula.metadata.get_serialization_data(),
            "config": self.formula.config.get_serialization_data(),
            "sha256": self.formula.get_sha256(),
        })

    @inlineCallbacks
    def test_get_formula(self):
        """
        A FormulaState should be available if one get()s a formula
        that was previously added into the manager.
        """
        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        formula_state = yield self.formula_state_manager.get_formula_state(
            "namespace:dummy-1")
        self.assertEquals(formula_state.id, "namespace:dummy-1")

    @inlineCallbacks
    def test_formula_state_attributes(self):
        """
        Verify that the basic (invariant) attributes of the
        FormulaState are correctly in place.
        """
        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        formula_state = yield self.formula_state_manager.get_formula_state(
            "namespace:dummy-1")
        self.assertEquals(formula_state.namespace, "namespace")
        self.assertEquals(formula_state.name, "dummy")
        self.assertEquals(formula_state.revision, 1)
        self.assertEquals(formula_state.id, "namespace:dummy-1")

    @inlineCallbacks
    def test_formula_state_metadata(self):
        """
        Check that the formula metadata was correctly saved and loaded.
        """
        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        formula_state = yield self.formula_state_manager.get_formula_state(
            "namespace:dummy-1")
        metadata = yield formula_state.get_metadata()
        self.assertEquals(metadata.name, "dummy")
        self.assertEquals(metadata.revision, 1)

    @inlineCallbacks
    def test_formula_state_config_options(self):
        """Verify ConfigOptions present and correct."""
        from ensemble.formula.tests.test_config import sample_yaml_data

        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        formula_state = yield self.formula_state_manager.get_formula_state(
            "namespace:dummy-1")
        config = yield formula_state.get_config()
        self.assertEquals(config.get_serialization_data(),
                          sample_yaml_data)



    @inlineCallbacks
    def test_get_non_existing_formula_prior_to_initialization(self):
        """
        Getting a formula before the formulas node was even
        initialized should raise an error about the formula not
        being present.
        """
        try:
            yield self.formula_state_manager.get_formula_state(
                "namespace:dummy-1")
        except FormulaStateNotFound, e:
            self.assertEquals(e.formula_id, "namespace:dummy-1")
        else:
            self.fail("Error not raised.")

    @inlineCallbacks
    def test_get_non_existing_formula(self):
        """
        Trying to retrieve a formula from the state when it was
        never added should raise an error.
        """
        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        try:
            yield self.formula_state_manager.get_formula_state(
                "anotherns:dummy-1")
        except FormulaStateNotFound, e:
            self.assertEquals(e.formula_id, "anotherns:dummy-1")
        else:
            self.fail("Error not raised.")

    @inlineCallbacks
    def test_get_sha256(self):
        """
        We should be able to retrieve the sha256 of a stored
        formula.
        """
        yield self.formula_state_manager.add_formula_state(
            "namespace", self.formula)
        formula_state = yield self.formula_state_manager.get_formula_state(
            "namespace:dummy-1")
        sha256 = yield formula_state.get_sha256()
        self.assertEquals(sha256, self.formula.get_sha256())
