import base64
import yaml
import zookeeper

from twisted.internet.defer import inlineCallbacks, succeed, returnValue

from ensemble.state.auth import make_identity, make_ace
from ensemble.state.errors import (
    StateChanged, StateNotFound, PrincipalNotFound)
from ensemble.state.security import (
    ACL, Principal, GroupPrincipal, OTPPrincipal, TokenDatabase,
    SecurityPolicy, SecureConnectionMixin,
    get_domain_security, apply_rules)

from ensemble.lib.testing import TestCase
from ensemble.state.tests.common import StateTestBase, security_test
from ensemble.tests.common import get_test_zookeeper_address

from txzookeeper.client import ZookeeperClient, ZOO_OPEN_ACL_UNSAFE

from txzookeeper.tests.utils import deleteTree


class PrincipalTests(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    def test_name(self):
        """Principals have names."""
        principal = Principal("foobar", "secret")
        self.assertEqual(principal.name, "foobar")

    def test_get_token(self):
        """An identity token can be gotten from a Principal."""
        principal = Principal("foobar", "secret")
        self.assertEqual(principal.get_token(),
                         make_identity("foobar:secret"))

    @inlineCallbacks
    def test_activate(self):
        """A principal can be used with a client connection."""
        client = yield self.get_zookeeper_client().connect()
        self.addCleanup(lambda: client.close())
        admin_credentials = "admin:admin"
        test_credentials = "test:test"
        auth_deferred = self.client.add_auth("digest", admin_credentials)

        acl = [make_ace(make_identity(admin_credentials), all=True),
               make_ace(make_identity(
                   test_credentials), read=True, create=True)]

        yield client.create("/acl-test", "content", acls=acl)

        # Wait on the auth after the create due to 770
        yield auth_deferred

        # Verify the acl is active
        yield self.assertFailure(
            client.get("/acl-test"), zookeeper.NoAuthException)

        # Attach the principal to the connection
        principal = Principal("test", "test")

        yield principal.attach(client)
        content, stat = yield client.get("/acl-test")
        self.assertEqual(content, "content")


class GroupPrincipalTests(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    def test_uninitialized_usage(self):
        """Attempting to access the name before initialized raises an error"""
        principal = GroupPrincipal(self.client, "/group-a")
        try:
            principal.name
        except ValueError:
            pass
        else:
            self.fail("Uninitialized usage should raise error")

    @inlineCallbacks
    def test_create(self):
        """An identity token can be gotten from a Principal."""
        principal = GroupPrincipal(self.client, "/group-a")
        yield principal.create("group/a", "zebra")
        self.assertEqual(principal.name, "group/a")
        yield self.assertFailure(
            principal.create("group/a", "zebra"),
            ValueError)

    @inlineCallbacks
    def test_initialize(self):
        principal = GroupPrincipal(self.client, "/group-a")
        yield principal.create("group/a", "zebra")

        principal = GroupPrincipal(self.client, "/group-a")
        yield principal.initialize()
        self.assertEqual(principal.name, "group/a")

    @inlineCallbacks
    def test_get_token(self):
        """An identity token can be gotten from a Principal."""
        principal = GroupPrincipal(self.client, "/group-a")
        yield principal.create("foobar", "secret")
        self.assertEqual(principal.get_token(),
                         make_identity("foobar:secret"))

    @inlineCallbacks
    def test_add_member(self):
        group = GroupPrincipal(self.client, "/group-a")
        yield group.create("group/a", "zebra")

        principal = Principal("aladdin", "genie")
        yield group.add_member(principal)
        acl, stat = yield self.client.get_acl("/group-a")
        self.assertEqual(
            acl[1:],
            [make_ace(principal.get_token(), read=True)])
        yield self.assertFailure(group.add_member(principal), StateChanged)

    @inlineCallbacks
    def test_get_members(self):
        group = GroupPrincipal(self.client, "/group-a")
        yield group.create("group/a", "zebra")

        principals = [
            Principal("aladdin", "genie"),
            Principal("babar", "elephant")]

        for p in principals:
            yield group.add_member(p)

        # Policy isn't active by default
        principals.insert(0, Principal("anyone", ""))

        members = yield group.get_members()
        self.assertEqual(
            [p.name for p in principals],
            members)

    @inlineCallbacks
    def test_remove_member(self):
        group = GroupPrincipal(self.client, "/group-a")
        yield group.create("group/a", "zebra")

        principal = Principal("aladdin", "genie")
        yield group.add_member(principal)
        yield group.remove_member(principal.name)
        acl, stat = yield self.client.get_acl("/group-a")
        self.assertEqual(acl[1:], [])

    @inlineCallbacks
    def test_activate(self):
        """A principal can be used with a client connection."""
        client = yield self.get_zookeeper_client().connect()
        self.addCleanup(lambda: client.close())

        admin_credentials = "admin:admin"
        test_credentials = "test:test"
        self.client.add_auth("digest", admin_credentials)

        acl = [make_ace(make_identity(admin_credentials), all=True),
               make_ace(make_identity(
                   test_credentials), read=True, create=True)]

        yield client.create("/acl-test", "content", acls=acl)

        # Verify the acl is active
        yield self.assertFailure(
            client.get("/acl-test"), zookeeper.NoAuthException)

        # Attach the principal to the connection
        group = GroupPrincipal(self.client, "/group-b")
        yield group.create("test", "test")
        yield group.attach(client)
        content, stat = yield client.get("/acl-test")
        self.assertEqual(content, "content")


class OTPPrincipalTests(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()
        self.client.create("/otp")

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    def set_otp_test_ace(self, test_ace=ZOO_OPEN_ACL_UNSAFE):
        """Set an additional OTP ACL entry for test cleanup."""
        OTPPrincipal.set_additional_otp_ace(test_ace)
        self.addCleanup(lambda: OTPPrincipal.set_additional_otp_ace(None))

    def test_using_uncreated_raises(self):
        """Principals have names."""
        principal = OTPPrincipal(self.client)
        try:
            principal.name
        except ValueError:
            pass
        else:
            self.fail("Use of an uncreated otp principal should raise error.")

    @inlineCallbacks
    def test_get_token(self):
        """An identity token can be gotten from a OTPPrincipal.

        The token returned is that of the stored credentials, not
        the serialized one time password principal.
        """
        self.set_otp_test_ace()

        principal = OTPPrincipal(self.client)
        yield principal.create("foobar", "secret")
        self.assertEqual(principal.get_token(),
                         make_identity("foobar:secret"))
        self.assertEqual(principal.name, "foobar")

    @inlineCallbacks
    def test_create(self):
        """A principal can be used with a client connection."""
        self.set_otp_test_ace()

        principal = OTPPrincipal(self.client)
        yield principal.create("foobar", "secret")

        children = yield self.client.get_children("/otp")
        self.assertEqual(len(children), 1)
        otp_path = "/otp/%s" % (children.pop())

        data, stat = yield self.client.get(otp_path)

        credentials = yaml.load(data)
        self.assertEqual(credentials["name"], "foobar")
        self.assertEqual(credentials["password"], "secret")

        acl, stat = yield self.client.get_acl(otp_path)
        self.assertEqual(len(acl), 2)

    @inlineCallbacks
    def test_serialize(self):
        """The principal can be serialized to just the otp data."""
        self.set_otp_test_ace()

        principal = OTPPrincipal(self.client)
        yield principal.create("foobar", "secret")

        otp_data = principal.serialize()
        path, user, password = base64.b64decode(otp_data).split(":")
        acl, stat = yield self.client.get_acl(path)

        self.assertEqual(principal.get_token(),
                         make_identity("foobar:secret"))
        self.assertEqual(principal.name, "foobar")

    @inlineCallbacks
    def test_consume(self):
        """The OTP serialization can be used to retrievethe actual credentials.
        """
        principal = OTPPrincipal(self.client)
        yield principal.create("foobar", "secret")
        otp_data = principal.serialize()
        path, _ = base64.b64decode(otp_data).split(":", 1)
        acl, stat = yield self.client.get_acl(path)

        # Verify that the OTP data is secure
        yield self.assertFailure(
            self.client.get(path), zookeeper.NoAuthException)

        name, password = yield OTPPrincipal.consume(self.client, otp_data)
        self.assertEqual(name, "foobar")
        self.assertEqual(password, "secret")
        children = yield self.client.get_children("/otp")
        self.assertFalse(children)


class TokenDatabaseTest(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()
        self.db = TokenDatabase(self.client, "/token-test")

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    @inlineCallbacks
    def test_add(self):
        principal = Principal("zebra", "zoo")
        yield self.db.add(principal)
        content, stat = yield self.client.get("/token-test")
        data = yaml.load(content)
        self.assertEqual(data, {"zebra": principal.get_token()})

    @inlineCallbacks
    def test_remove(self):
        principal = Principal("zebra", "zoo")
        yield self.db.add(principal)
        yield self.db.remove(principal)
        content, stat = yield self.client.get("/token-test")
        data = yaml.load(content)
        self.assertEqual(data, {"zebra": principal.get_token()})

    @inlineCallbacks
    def test_get(self):
        principal = Principal("zebra", "zoo")
        yield self.db.add(principal)
        token = yield self.db.get(principal.name)
        self.assertEqual(token, principal.get_token())

    @inlineCallbacks
    def test_get_nonexistant(self):
        principal = Principal("zebra", "zoo")
        error = yield self.assertFailure(self.db.get(principal.name),
                                   PrincipalNotFound)
        self.assertIn("zebra", str(error))


class PolicyTest(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()
        self.tokens = TokenDatabase(self.client)
        yield self.tokens.add(Principal("admin", "admin"))
        self.policy = SecurityPolicy(self.client, self.tokens)

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    def test_client_property_access(self):
        self.assertTrue(self.policy.client is self.client)

    @inlineCallbacks
    def test_get_token(self):
        self.assertEqual(
            (yield self.policy.get_token("admin")),
            Principal("admin", "admin").get_token())

    @inlineCallbacks
    def test_default_no_owner_no_rules_gives_admin_access(self):
        """By default the policy setups a global access for the cli admins.
        """
        acl = yield self.policy("/random")
        self.assertIn(
            make_ace(Principal("admin", "admin").get_token(), all=True), acl)

    @inlineCallbacks
    def test_default_no_rules_gives_global_access(self):
        """If no rules match, the default acl gives authenticated users access.

        XXX/TODO: This is intended as a temporary crutch for
        integration of the security machinery, not a long term
        solution.
        """
        acl = yield self.policy("/random")
        self.assertIn(make_ace("anyone", "world", all=True), acl)

    @inlineCallbacks
    def test_rule_match_suppress_open_access(self):
        """If a rule returns an acl, then no default access is given."""
        principal = Principal("foobar", "foobar")
        self.policy.add_rule(lambda policy, path: [
            make_ace(principal.get_token(), all=True)])
        acl = yield self.policy("/random")
        # Check for matched rule ACL
        self.assertIn(make_ace(principal.get_token(), all=True), acl)
        # Verify no default access
        self.assertNotIn(make_ace("anyone", "world", all=True), acl)

    @inlineCallbacks
    def test_rule_that_returns_deferred(self):
        """If a rule may do additional lookups, resulting in deferred values.
        """
        principal = Principal("foobar", "foobar")
        self.policy.add_rule(lambda policy, path: succeed([
            make_ace(principal.get_token(), all=True)]))
        acl = yield self.policy("/random")
        # Check for matched rule ACL
        self.assertIn(make_ace(principal.get_token(), all=True), acl)
        # Verify no default access
        self.assertNotIn(make_ace("anyone", "world", all=True), acl)

    @inlineCallbacks
    def test_owner_ace(self):
        """If an owner is set, all nodes ACLs will have an owner ACE.
        """
        owner = Principal("john", "doe")
        self.policy.set_owner(owner)
        acl = yield self.policy("/random")
        self.assertIn(make_ace(owner.get_token(), all=True), acl)


class SecureConnection(SecureConnectionMixin, ZookeeperClient):
    pass


class SecureConnectionTest(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield SecureConnection(
            get_test_zookeeper_address()).connect()

        admin = Principal("admin", "admin")

        self.token_db = TokenDatabase(self.client)
        yield self.token_db.add(admin)
        self.policy = SecurityPolicy(self.client, self.token_db, owner=admin)
        yield admin.attach(self.client)

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    @inlineCallbacks
    def test_create_without_policy(self):
        """If no policy is set the connection behaves normally"""
        def rule(policy, path):
            return [make_ace(Principal("magic", "not").get_token(), all=True)]

        self.policy.add_rule(rule)
        yield self.client.create("/xyz")
        acl, stat = yield self.client.get_acl("/xyz")
        self.assertEqual(acl, [ZOO_OPEN_ACL_UNSAFE])

    @inlineCallbacks
    def test_create_with_policy(self):
        """If a policy is set ACL are determined by the policy."""

        def rule(policy, path):
            return [make_ace(Principal("magic", "not").get_token(), all=True)]

        self.policy.add_rule(rule)
        self.client.set_security_policy(self.policy)

        yield self.client.create("/xyz")
        acl, stat = yield self.client.get_acl("/xyz")

        self.assertEqual(
            acl,
            [make_ace(Principal("magic", "not").get_token(), all=True),
             make_ace(Principal("admin", "admin").get_token(), all=True)])


class ACLTest(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = yield self.get_zookeeper_client().connect()
        self.tokens = TokenDatabase(self.client)
        self.admin = Principal("admin", "admin")
        yield self.tokens.add(self.admin)
        self.policy = SecurityPolicy(self.client, self.tokens)
        yield self.admin.attach(self.client)

    def tearDown(self):
        deleteTree(handle=self.client.handle)
        self.client.close()

    @inlineCallbacks
    def test_acl_on_non_existant_node(self):
        acl = ACL(self.client, "abc")
        yield self.assertFailure(acl.grant("admin", all=True), StateNotFound)

    @inlineCallbacks
    def test_acl_without_admin(self):
        """A client needs an attached principle with the admin perm to set acl.
        """
        client = yield self.get_zookeeper_client().connect()
        principal = Principal("zebra", "stripes")
        yield self.tokens.add(principal)
        yield principal.attach(client)
        yield self.client.create(
            "/abc",
            acls=[make_ace(self.admin.get_token(), all=True)])

        acl = ACL(client, "/abc")
        yield self.assertFailure(
            acl.grant("zebra", all=True),
            zookeeper.NoAuthException)

    @inlineCallbacks
    def test_grant(self):
        path = yield self.client.create("/abc")
        acl = ACL(self.client, "/abc")
        yield acl.grant("admin", all=True)
        node_acl, stat = yield self.client.get_acl(path)
        self.assertEqual(
            node_acl,
            [ZOO_OPEN_ACL_UNSAFE,
             make_ace(self.admin.get_token(), all=True)])

    @inlineCallbacks
    def test_grant_not_in_token_database(self):
        path = yield self.client.create("/abc")
        acl = ACL(self.client, path)
        yield self.assertFailure(acl.grant("zebra"), PrincipalNotFound)

    @inlineCallbacks
    def test_grant_with_existing_ace(self):
        path = yield self.client.create(
            "/abc",
            acls=[
                make_ace(self.admin.get_token(),
                         all=True)]
            )

        acl = ACL(self.client, path)
        yield self.assertFailure(
            acl.grant("admin", admin=True),
            StateChanged)

    @inlineCallbacks
    def test_concurrent_conflicting_grant(self):
        path = yield self.client.create("/abc", acls=[
            make_ace(self.admin.get_token(), all=True)])
        principal = Principal("zebra", "stripes")
        yield self.tokens.add(principal)

        @inlineCallbacks
        def modify_post_get(path):
            acl, stat = yield self.client.get_acl(path)
            acl_copy = list(acl)
            acl_copy.append(make_ace(principal.get_token(), read=True))
            yield self.client.set_acl(path, acl_copy)
            returnValue((acl, stat))

        mock_client = self.mocker.proxy(self.client)
        mock_client.get("/auth-tokens")
        self.mocker.passthrough()

        mock_client.get_acl(path)
        self.mocker.call(modify_post_get)

        mock_client.set_acl(
            path,
            [make_ace(self.admin.get_token(), all=True),
             make_ace(principal.get_token(), write=True)],
            version=0)
        self.mocker.passthrough()

        mock_client.get_acl(path)
        self.mocker.passthrough()
        self.mocker.replay()

        acl = ACL(mock_client, path)
        yield self.assertFailure(
            acl.grant(principal.name, write=True),
            StateChanged)

        self.client.get_acl(path)

    @inlineCallbacks
    def test_prohibit(self):
        principal = Principal("zebra", "stripes")
        yield self.tokens.add(principal)

        path = yield self.client.create("/abc", acls=[
            make_ace(self.admin.get_token(), all=True),
            make_ace(principal.get_token(), write=True)])

        acl = ACL(self.client, path)
        yield acl.prohibit("zebra")

        acl, stat = yield self.client.get_acl(path)
        self.assertEqual(
            acl, [make_ace(self.admin.get_token(), all=True)])

    @inlineCallbacks
    def test_prohibit_non_existant_node(self):
        acl = ACL(self.client, "/abc")
        yield self.assertFailure(
            acl.prohibit("zebra"), StateNotFound)

    @inlineCallbacks
    def test_prohibit_not_in_acl(self):
        principal = Principal("zebra", "stripes")
        yield self.tokens.add(principal)

        path = yield self.client.create("/abc", acls=[
            make_ace(self.admin.get_token(), all=True)])

        acl = ACL(self.client, path)
        # We get to the same end state so its fine.
        yield acl.prohibit("zebra")

        acl, stat = yield self.client.get_acl(path)
        self.assertEqual(
            acl, [make_ace(self.admin.get_token(), all=True)])


class SecurityModuleAPITest(StateTestBase):
    """
    Test of the top level conviencce functions in the security module.
    """

    @security_test
    @inlineCallbacks
    def test_apply_rules(self):
        """Calculates and sets an acl using the security rules.
        """
        cli_group = Principal("admin", "admin")
        yield cli_group.attach(self.client)
        yield TokenDatabase(self.client).add(cli_group)

        principal = Principal("testadmin", "testadmin")
        token = principal.get_token()

        def get_test_rules():
            return [lambda x, y: [make_ace(token, all=True)]]

        get_default_rules = self.mocker.replace(
            "ensemble.state.securityrules.get_default_rules")

        get_default_rules()
        self.mocker.call(get_test_rules)
        self.mocker.replay()

        path = yield self.client.create("/moon")
        yield apply_rules(self.client, path)
        acl, stat = yield self.client.get_acl(path)
        self.assertEqual(
            acl,
            [make_ace(token, all=True),
             make_ace(cli_group.get_token(), all=True)])


class DomainSecurityTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(DomainSecurityTest, self).setUp()
        path = yield self.client.create("/zoo")
        self.security = get_domain_security(self.client, path)

    @inlineCallbacks
    def test_add_group(self):
        """Creates a group principal and adds it to the token db.
        """
        self.security.enabled = False
        group = yield self.security.add_group("group-a")
        self.assertEqual(group, None)
        self.assertFalse((yield self.client.exists("/zoo/group")))

        self.security.enabled = True
        group = yield self.security.add_group("group-a")
        self.assertTrue(isinstance(group, GroupPrincipal))
        token = yield TokenDatabase(self.client).get("group-a")
        self.assertEqual(token, group.get_token())

    @security_test
    @inlineCallbacks
    def test_consume_otp(self):
        """OTP can be consumed to retrieve the persistent principal."""

        yield self.assertFailure(
            self.security.consume_otp(), PrincipalNotFound)

        self.security.enabled = False
        self.assertEqual((yield self.security.consume_otp()), None)

        self.security.enabled = True
        yield self.security.apply_otp("aladdin")
        principal = yield self.security.consume_otp()
        self.assertEqual(principal.name, "aladdin")
        token = yield TokenDatabase(self.client).get("aladdin")
        self.assertEqual(principal.get_token(), token)
        self.assertFalse(
            (yield self.client.get_children("/otp")))

    @security_test
    @inlineCallbacks
    def test_apply_otp(self):
        """Creates an otp principal and stores it serialization at path.
        """
        self.security.enabled = False
        yield self.security.apply_otp("aladdin")

        # Verify the otp serialization on the node
        contents, stat = yield self.client.get("/zoo")
        data = yaml.load(contents)
        self.assertFalse(data)

        self.security.enabled = True
        yield self.security.apply_otp("aladdin")

        # Verify the otp serialization on the node
        contents, stat = yield self.client.get("/zoo")
        data = yaml.load(contents)
        self.assertIn("otp-identity", data)

        # Compare the persitsent user generated token to that in the tokendb
        user, password = yield OTPPrincipal.consume(
            self.client, data["otp-identity"])
        token = yield TokenDatabase(self.client).get("aladdin")
        self.assertEqual(
            token, make_identity("%s:%s" % (user, password)))

    @inlineCallbacks
    def test_add_remove_group_member(self):
        """Adding and removing members of a security group."""
        self.security.enabled = True
        group = yield self.security.add_group("group-a")

        self.security.enabled = False
        yield self.security.add_group_member(Principal("music", "match"))
        members = yield group.get_members()
        self.assertNotIn("music", members)

        self.security.enabled = True
        yield self.security.add_group_member(Principal("music", "match"))
        members = yield group.get_members()
        self.assertIn("music", members)

        self.security.enabled = False
        yield self.security.remove_group_member("music")
        members = yield group.get_members()
        self.assertIn("music", members)

        self.security.enabled = True
        yield self.security.remove_group_member("music")
        members = yield group.get_members()
        self.assertNotIn("music", members)
