from twisted.internet.defer import inlineCallbacks

from ensemble.state.auth import make_ace
from ensemble.state.security import Principal, TokenDatabase, SecurityPolicy
from ensemble.state import securityrules as rules
from ensemble.state.topology import InternalTopology

from ensemble.state.tests.common import StateTestBase


class RulesTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(RulesTest, self).setUp()
        self.principals = {
            "admin": Principal("admin", "admin"),
            "provider/0": Principal("provider/0", "provisioning-agent"),
            "ensemble-provider": Principal(
                "ensemble-provider", "provisioning-group"),
            "machine/0": Principal("machine/0", "machine-agent"),
            "service/0": Principal("service/0", "unit-agent"),
            "service": Principal("service", "service-group")
            }
        self.token_db = TokenDatabase(self.client)

        for p in self.principals.values():
            yield self.token_db.add(p)

        self.policy = SecurityPolicy(self.client, self.token_db)

    @inlineCallbacks
    def test_formula_rule(self):
        self.assertEqual(
            (yield rules.formula_rule(self.policy, "/abc")),
            None)
        self.assertEqual(
            (yield rules.formula_rule(self.policy, "/formulas")),
            [make_ace("anyone", "world", read=True)])
        self.assertEqual(
            (yield rules.formula_rule(self.policy, "/formulas/formula-0")),
            [make_ace("anyone", "world", read=True)])

    @inlineCallbacks
    def test_machine_rule(self):
        self.assertEqual(
            (yield rules.machine_rule(self.policy, "/xyz")),
            None)
        self.assertEqual(
            (yield rules.machine_rule(self.policy, "/machines")),
            [make_ace(
                (yield self.token_db.get("ensemble-provider")),
                read=True)])
        self.assertEqual(
            (yield rules.machine_rule(self.policy, "/machines/machine-1")),
            [make_ace(
                (yield self.token_db.get("ensemble-provider")),
                read=True, admin=True)])

    @inlineCallbacks
    def test_relation_rule_client_server(self):
        topology = InternalTopology()
        topology.add_service("service-0000001", "mysql")
        topology.add_service("service-0000002", "wordpress")
        topology.add_relation("relation-0000001", "database")
        topology.assign_service_to_relation(
            "relation-0000001", "service-0000001", "app", "server")
        topology.assign_service_to_relation(
            "relation-0000001", "service-0000002", "db", "client")

        mysql_principal = Principal("mysql", "mysql")
        wordpress_principal = Principal("wordpress", "wordpress")
        yield self.token_db.add(mysql_principal)
        yield self.token_db.add(wordpress_principal)

        yield self.client.create("/topology", topology.dump())

        self.assertEqual(
            (yield rules.relation_rule(self.policy, "/xyz")),
            None)

        self.assertEqual(
            (yield rules.relation_rule(self.policy, "/relations")),
            None)

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/settings/unit-x")),
            [make_ace(mysql_principal.get_token(), read=True),
             make_ace(wordpress_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/client/unit-x")),
            [make_ace(mysql_principal.get_token(), read=True),
             make_ace(wordpress_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/server")),
            [make_ace(mysql_principal.get_token(), create=True, read=True),
             make_ace(wordpress_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/client")),
            [make_ace(wordpress_principal.get_token(), create=True, read=True),
             make_ace(mysql_principal.get_token(), read=True)]
            )

    @inlineCallbacks
    def test_relation_rule_peer(self):
        topology = InternalTopology()
        topology.add_service("service-0000001", "riak")
        topology.add_relation("relation-0000001", "ring")
        topology.assign_service_to_relation(
            "relation-0000001", "service-0000001", "ring", "peer")
        riak_principal = Principal("riak", "riak")
        yield self.token_db.add(riak_principal)
        yield self.client.create("/topology", topology.dump())

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/settings/unit-x")),
            [make_ace(riak_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/peer/unit-x")),
            [make_ace(riak_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.relation_rule(
                self.policy, "/relations/relation-0000001/peer")),
            [make_ace(riak_principal.get_token(), create=True, read=True)]
            )

    @inlineCallbacks
    def test_global_rule(self):
        self.assertEqual(
            (yield rules.global_rule(self.policy, "/xyz")),
            None)
        self.assertEqual(
            (yield rules.global_rule(self.policy, "/topology")),
            [make_ace("auth", "world", read=True)])

        self.assertEqual(
            (yield rules.global_rule(self.policy, "/auth-tokens")),
            [make_ace("auth", "world", write=True, read=True)])

    @inlineCallbacks
    def test_service_rule(self):

        topology = InternalTopology()
        topology.add_service("service-0000001", "riak")
        topology.add_service_unit("service-0000001", "unit-0000001")
        riak_principal = Principal("riak", "riak")
        self.token_db.add(riak_principal)
        yield self.client.create("/topology", topology.dump())

        self.assertEqual(
            (yield rules.service_rule(self.policy, "/abc")),
            None)

        self.assertEqual(
            (yield rules.service_rule(self.policy, "/services")),
            None)

        self.assertEqual(
            (yield rules.service_rule(
                self.policy, "/services/service-0000001")),
            None,
            )

        self.assertEqual(
            (yield rules.service_rule(
                self.policy, "/services/service-0000001/config")),
            [make_ace(
                riak_principal.get_token(), read=True)]
            )

        self.assertEqual(
            (yield rules.service_rule(
                self.policy, "/services/service-0000001/exposed")),
            [make_ace(
                self.principals["ensemble-provider"].get_token(), read=True)]
            )

    @inlineCallbacks
    def test_unit_rule(self):

        topology = InternalTopology()
        topology.add_service("service-0000001", "riak")
        topology.add_service_unit("service-0000001", "unit-0000001")
        yield self.client.create("/topology", topology.dump())

        riak_unit = Principal("riak/0", "abc")
        yield self.token_db.add(riak_unit)

        self.assertEqual(
            (yield rules.unit_rule(self.policy, "/xyz")), None)

        self.assertEqual(
            (yield rules.unit_rule(self.policy, "/units")), None)

        self.assertEqual(
            (yield rules.unit_rule(self.policy, "/units/unit-0000001")),
            [make_ace(riak_unit.get_token(),
                      read=True, write=True, delete=True, create=True)])

        self.assertEqual(
            (yield rules.unit_rule(self.policy, "/units/unit-0000001/ports")),
            [make_ace(
                self.principals["ensemble-provider"].get_token(), read=True)])

        self.assertEqual(
            (yield rules.unit_rule(self.policy, "/units/unit-0000001/debug")),
            [make_ace(riak_unit.get_token(), read=True)])
