import os

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.formula import get_formula_from_path
from ensemble.state.errors import FormulaStateNotFound
from ensemble.formula.bundle import FormulaBundle
from ensemble.formula.tests.test_directory import sample_directory
from ensemble.formula.publisher import FormulaPublisher
from ensemble.environment.config import EnvironmentsConfig
from ensemble.state.environment import EnvironmentStateManager

from ensemble.state.tests.common import StateTestBase

from ensemble.unit.formula import download_formula

SAMPLE_ENV = """\
ensemble: environments

environments:
  myfirstenv:
    type: dummy
    foo: bar
    storage-directory: %s
"""


class FormulaPublisherTestBase(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(FormulaPublisherTestBase, self).setUp()

        config = self.get_test_environment_config()
        environment = config.get_default()

        # Store the environment to zookeeper
        environment_state_manager = EnvironmentStateManager(self.client)
        yield environment_state_manager.set_config_state(config, "myfirstenv")

        # Load the environment
        self.provider = environment.get_machine_provider()
        self.storage = self.provider.get_file_storage()

    def get_test_environment_config(self):
        sample_config = SAMPLE_ENV % self.makeDir()
        config = EnvironmentsConfig()
        config.parse(sample_config)
        return config

    @inlineCallbacks
    def publish_formula(self, formula_path=sample_directory):
        formula = get_formula_from_path(formula_path)
        publisher = FormulaPublisher(self.client, self.storage)
        yield publisher.add_formula("local", formula)
        formula_states = yield publisher.publish()
        returnValue((formula, formula_states[0]))


class DownloadTestCase(FormulaPublisherTestBase):

    @inlineCallbacks
    def test_formula_download(self):
        """Downloading a formula should store the formula locally.
        """
        formula, formula_state = yield self.publish_formula()
        formula_directory = self.makeDir()

        # Download the formula
        yield download_formula(
            self.client, formula_state.id, formula_directory)

        # Verify the downloaded copy
        checksum = formula.get_sha256()
        formula_key = "%s:%s" % (
            formula.metadata.get_formula_id("local"), checksum)
        formula_path = os.path.join(formula_directory, formula_key)

        self.assertTrue(os.path.exists(formula_path))
        bundle = FormulaBundle(formula_path)
        self.assertTrue(
            bundle.metadata.get_formula_id("local"),
            formula.metadata.get_formula_id("local"))
        self.assertEqual(checksum, bundle.get_sha256())

    @inlineCallbacks
    def test_formula_download_not_found(self):
        """An error is raised if trying to download a non existant formula.
        """
        formula_directory = self.makeDir()

        # Download the formula
        error = yield self.assertFailure(
            download_formula(
                self.client, "local:mickey-21", formula_directory),
            FormulaStateNotFound)

        self.assertIn("Formula 'local:mickey-21' was not found", str(error))
